"""
Topographica cortical map simulator package.

Topographica is designed as a collection of packages from which
elements can be selected to model specific systems.  For more
information, see the individual subpackages::

  base           - Core Topographica functions and classes
  plotting       - Visualization functions and classes
  analysis       - Analysis functions and classes (besides plotting)
  tkgui          - Tk-based graphical user interface (GUI)
  command        - High-level user commands
  misc           - Various useful independent modules

The Topographica primitives library consists of a family of classes
that can be used with the above functions and classes::

  sheet          - Sheet classes: 2D arrays of processing units
  projection     - Projection classes: connections between Sheets
  pattern        - PatternGenerator classes: 2D input or weight patterns 
  ep             - EventProcessor classes: other simulation objects
  transferfn     - Transfer functions, for e.g. normalization or squashing
  responsefn     - Calculate the response of a Projection
  learningfn     - Adjust weights for a Projection
  coordmapper    - CoordinateMapperFn classes: map coords between Sheets

Each of the library directories can be extended with new classes of
the appropriate type, just by adding a new .py file to that directory.
E.g. new PatternGenerator classes can be added to pattern/, and will
then show up in the GUI menus as potential input patterns.

$Id$
"""

# The tests and the GUI are omitted from this list, and have to be
# imported explicitly if desired.
__all__ = ['analysis',
           'base',
           'command',
           'coordmapper',
           'ep',
           'learningfn',
           'misc',
           'numbergen',
           'transferfn',
           'pattern',
           'plotting',
           'projection',
           'responsefn',
           'sheet']


# The version number will eventually need to be set automatically from git
release='0.9.8-1'
__version__=release
version='12131'

import param 
import os
import errno
import platform


def _win_documents_path():
    """
    Return the Windows "My Documents" folder path, if available.
    """
    # Accesses the Windows API via ctypes
    import ctypes
    import ctypes.wintypes

    CSIDL_PERSONAL = 0x0005
    dll = ctypes.windll.shell32
    buf = ctypes.create_unicode_buffer(ctypes.wintypes.MAX_PATH + 1)
    if dll.SHGetSpecialFolderPathW(None, buf, CSIDL_PERSONAL, False):
        return buf.value.encode()
    else:
        raise ValueError


def _xdg_documents_path():
    """
    Return the Linux/UNIX XDG "Documents" folder path, if available.
    """
    # Runs the xdg-user-dir command from xdg-utils
    # (which comes with most Linux systems)

    import subprocess
    p = subprocess.Popen(["xdg-user-dir", "DOCUMENTS"], stdout=subprocess.PIPE)
    path = p.communicate()[0].strip()
    if path:
        return path
    else:
        raise ValueError


# Determine the appropriate location in which to create files
# on this operating system
_default_output_path = os.path.join(os.path.expanduser('~'), 
                                    'Documents', 'Topographica')
try:
    documents = _xdg_documents_path()
    _default_output_path = os.path.join(documents, 'Topographica')
except: pass

if platform.system() == 'Windows':
    try:
        documents = _win_documents_path()
        _default_output_path = os.path.join(documents, 'Topographica')
    except: pass


# Make sure the default output path exists
if not os.path.exists(_default_output_path):
    print "Creating %s"%_default_output_path
    try:
        os.makedirs(_default_output_path)
    except OSError, e:
        if e.errno != errno.EEXIST:
            raise


# Location of topo/ package. This kind of thing won't work with py2exe
# etc. Need to see if we can get rid of it.
_package_path = os.path.split(__file__)[0]

param.normalize_path.prefix = _default_output_path
# CEBALERT: _default_output_path shouldn't be in there, right?
param.resolve_path.search_paths+=([_default_output_path] + [_package_path])



# CEBALERT (about PIL):
# PIL can be installed so that it's e.g. "from PIL import Image" or
# just "import Image".  The code below means Image etc are always
# imported, but then the rest of topographica can consistently use
# Image (rather than a try/except, such as the one below). An
# alternative would be an import hook, which would only run on
# attempting to import Image etc.

try:
    import Image
except ImportError:
    from PIL import Image, ImageOps, ImageDraw, ImageFont  # pyflakes:ignore (try/except import)
    import sys
    sys.modules['Image']=Image
    sys.modules['ImageOps']=ImageOps
    sys.modules['ImageDraw']=ImageDraw
    sys.modules['ImageFont']=ImageFont

# ImageTk is completely optional
try:
    import ImageTk
except ImportError:
    try:
        from PIL import ImageTk  # pyflakes:ignore (try/except import)
        import sys
        sys.modules['ImageTk']=ImageTk
    except ImportError:
        pass


# CEBALERT: can we move these pickle support functions elsewhere?  In
# fact, can we just gather all the non-legacy pickle garbage into one
# place? Pickle clutter adds complexity, and having all the pickle
# support in one places makes it easier for other developers to copy
# in new situations.

# (note that these _pickle_support functions also work for deep copying)

def _numpy_ufunc_pickle_support():
    """
    Allow instances of numpy.ufunc to pickle.
    """
    # Remove this when numpy.ufuncs themselves support pickling.
    # Code from Robert Kern; see:
    #http://news.gmane.org/find-root.php?group=gmane.comp.python.numeric.general&article=13400
    from numpy import ufunc
    import copy_reg

    def ufunc_pickler(ufunc):
        """Return the ufunc's name"""
        return ufunc.__name__

    copy_reg.pickle(ufunc,ufunc_pickler)

_numpy_ufunc_pickle_support()


def _mpq_pickle_support():
    """Allow instances of gmpy.mpq to pickle."""
    from gmpy import mpq
    mpq_type = type(mpq(1,10)) # gmpy doesn't appear to expose the type another way
    import copy_reg
    copy_reg.pickle(mpq_type,lambda q: (mpq,(q.digits(),)))



def _instance_method_pickle_support():
    """Allow instance methods to pickle."""
    # CB: well, it seems to work - maybe there are cases where this
    # wouldn't work?
    # Alternative technique (totally different approach), but would
    # only work with pickle (not cPickle):
    # http://code.activestate.com/recipes/572213/
    def _pickle_instance_method(mthd):
        mthd_name = mthd.im_func.__name__
        obj = mthd.im_self
        return getattr, (obj,mthd_name)

    import copy_reg, types
    copy_reg.pickle(types.MethodType, _pickle_instance_method)

_instance_method_pickle_support()


from topo.base.simulation import Simulation

# Set the default value of Simulation.time_type to gmpy.mpq. If gmpy
# is unavailable, use the slower fixedpoint.FixedPoint.
try:
    import gmpy
    Simulation.time_type = gmpy.mpq
    Simulation.time_type_args = ()
    _mpq_pickle_support()
except ImportError:
    import topo.misc.fixedpoint as fixedpoint
    param.Parameterized().warning('gmpy.mpq not available; using slower fixedpoint.FixedPoint for simulation time.')
    Simulation.time_type = fixedpoint.FixedPoint
    Simulation.time_type_args = (4,)  # gives precision=4

    # Provide a fake gmpy.mpq (to allow e.g. pickled test data to be
    # loaded).
    # CEBALERT: can we move this into whatever test needs it? I guess
    # it also has to be here to allow snapshots saved using gmpy time
    # type to open on systems where gmpy is not available.
    from topo.misc.util import gmpyImporter
    import sys
    sys.meta_path.append(gmpyImporter())
        


sim = Simulation() 


# numbergen used to be part of topo; import it there for backwards compatibility 
# and set the time function to be topo.sim.time()
import sys,numbergen
sys.modules['topo.numbergen']=numbergen
sys.modules['topo.numbergen.basic']=numbergen
numbergen.ExponentialDecay.time_fn = sim.time

# imagen used to be part of topo; import its files at their former locations
# for backwards compatibility and set the time function to be topo.sim.time()
import imagen as pattern
import imagen.random, imagen.image
sys.modules['topo.base.boundingregion']=pattern.boundingregion
sys.modules['topo.base.sheetcoords']=pattern.sheetcoords
sys.modules['topo.base.patterngenerator']=pattern.patterngenerator
sys.modules['topo.misc.patternfn']=pattern.patternfn
sys.modules['topo.pattern']=pattern
sys.modules['topo.pattern.basic']=pattern
sys.modules['topo.pattern.random']=pattern.random
sys.modules['topo.pattern.image']=pattern.image
sys.modules['topo.pattern.rds']=imagen.random
pattern.Translator.time_fn = sim.time



def about(display=True):
    """Print release and licensing information."""

    ABOUT_TEXT = """
Pre-release version %s (%s) of Topographica; an updated
version may be available from topographica.org.

This program is free, open-source software available under the BSD
license (http://www.opensource.org/licenses/bsd-license.php).
"""%(release,version)
    if display:
        print ABOUT_TEXT
    else:        
        return ABOUT_TEXT


# Set most floating-point errors to be fatal for safety; see
# topo/misc/patternfn.py for examples of how to disable
# the exceptions when doing so is safe.  Underflow is always
# considered safe; e.g. input patterns can be very small
# at large distances, and when they are scaled by small
# weights underflows are common and not a problem.
from numpy import seterr
old_seterr_settings=seterr(all="raise",under="ignore")




"""
Test cases for the numbergen module.

$Id$
"""

import unittest
from topo import numbergen


_seed = 0  # keep tests deterministic
_iterations = 1000


class TestUniformRandom(unittest.TestCase):
    def test_range(self):
        lbound = 2.0
        ubound = 5.0
        gen = numbergen.UniformRandom(
                seed=_seed,
                lbound=lbound,
                ubound=ubound)
        for _ in xrange(_iterations):
            value = gen()
            self.assertTrue(lbound <= value < ubound)

class TestUniformRandomOffset(unittest.TestCase):
    def test_range(self):
        lbound = 2.0
        ubound = 5.0
        gen = numbergen.UniformRandomOffset(
                seed=_seed,
                mean=(ubound + lbound) / 2,
                range=ubound - lbound)
        for _ in xrange(_iterations):
            value = gen()
            self.assertTrue(lbound <= value < ubound)


cases = [
            TestUniformRandom,
            TestUniformRandomOffset,
        ]

suite = unittest.TestSuite()

suite.addTests([unittest.makeSuite(case) for case in cases])

if __name__ == '__main__':
    unittest.TextTestRunner(verbosity=1).run(suite)



"""
FeatureResponses and associated functions and classes.

These classes implement map and tuning curve measurement based
on measuring responses while varying features of an input pattern.

$Id$
"""


import copy

from math import pi
from colorsys import hsv_to_rgb

import numpy
from numpy import zeros, empty, object_, size, vectorize, fromfunction
from numpy.oldnumeric import Float

import param
from param.parameterized import ParameterizedFunction, ParamOverrides

import topo
import topo.base.sheetcoords
from topo.base.arrayutil import wrap
from topo.base.cf import CFSheet
from topo.base.functionfamily import PatternDrivenAnalysis
from topo.base.sheet import Sheet, activity_type
from topo.base.sheetview import SheetView
from topo.command import pattern_present,restore_input_generators, save_input_generators
from topo.misc.distribution import Distribution, DistributionStatisticFn, DSF_MaxValue, DSF_WeightedAverage
from topo.misc.util import cross_product, frange
from topo import pattern
from topo.pattern import SineGrating, Gaussian, RawRectangle, Disk
from topo.plotting.plotgroup import plotgroups
from topo.sheet import GeneratorSheet


# CB: having a class called DistributionMatrix with an attribute
# distribution_matrix to hold the distribution matrix seems silly.
# Either rename distribution_matrix or make DistributionMatrix into
# a matrix.
class DistributionMatrix(param.Parameterized):
    """
    Maintains a matrix of Distributions (each of which is a dictionary
    of (feature value: activity) pairs).
    
    The matrix contains one Distribution for each unit in a
    rectangular matrix (given by the matrix_shape constructor
    argument).  The contents of each Distribution can be updated for a
    given bin value all at once by providing a matrix of new values to
    update().

    The results can then be accessed as a matrix of weighted averages
    (which can be used as a preference map) and/or a selectivity
    map (which measures the peakedness of each distribution).
    """
    def __init__(self,matrix_shape,axis_range=(0.0,1.0), cyclic=False,keep_peak=True):
        """Initialize the internal data structure: a matrix of Distribution objects."""
        self.axis_range=axis_range
        new_distribution = vectorize(lambda x: Distribution(axis_range,cyclic,keep_peak),
                                     doc="Return a Distribution instance for each element of x.")
        self.distribution_matrix = new_distribution(empty(matrix_shape))


    def update(self, new_values, bin):
        """Add a new matrix of histogram values for a given bin value."""
        ### JABHACKALERT!  The Distribution class should override +=,
        ### rather than + as used here, because this operation
        ### actually modifies the distribution_matrix, but that has
        ### not yet been done.  Alternatively, it could use a different
        ### function name altogether (e.g. update(x,y)).
        self.distribution_matrix + fromfunction(vectorize(lambda i,j: {bin:new_values[i,j]}),
                                                new_values.shape)  
        
    def apply_DSF( self, dsf ):
        """
        Apply the given dsf DistributionStatisticFn on each element of the distribution_matrix

        Return a dictionary of dictionaries, with the same structure of the called
        DistributionStatisticFn, but with matrices as values, instead of scalars
        """

        shape   = self.distribution_matrix.shape
        result  = {}

        # this is an extra call to the dsf() DistributionStatisticFn, in order to retrieve
        # the dictionaries structure, and allocate the necessary matrices
        r0      = dsf( self.distribution_matrix[ 0, 0 ] )
        for k, maps in r0.items():
            result[ k ] = {}
            for m in maps.keys():
                result[ k ][ m ]    = zeros( shape, Float )

        for i in range( shape[ 0 ] ):
            for j in range( shape[ 1 ] ):
                response        = dsf( self.distribution_matrix[ i, j ] )
                for k, d in response.items():
                    for item, item_value in d.items():
                        result[ k ][ item ][ i, j ]  = item_value

        return result



class FullMatrix(param.Parameterized):
    """
    Records the output of every unit in a sheet, for every combination of feature values.
    Useful for collecting data for later analysis while presenting many input patterns.
    """

    def __init__(self,matrix_shape,features):
        self.matrix_shape = matrix_shape
        self.features = features
        self.dimensions = ()
        for f in features:
            self.dimensions = self.dimensions + (size(f.values),)
        self.full_matrix = empty(self.dimensions,object_)


    def update(self, new_values, feature_value_permutation):
        """Add a new matrix of histogram values for a given bin value."""
        index = ()
        for f in self.features:
            for ff,value in feature_value_permutation:
                if(ff == f.name):
                    index = index + (f.values.index(value),)
        self.full_matrix[index] = new_values




# CB: FeatureResponses and ReverseCorrelation need cleanup; I began but haven't finished.
# JABALERT: At least:
# - Move features out of __init__ and into measure_responses
# - Change measure_responses to __call__, as it's the only thing this
#   class really does
# - Make the other methods private (_) since they are for internal use
# - Possibly -- make the __call__ methods have the same signature?
# - Clean up the inheritance hierarchy?


class FeatureResponses(PatternDrivenAnalysis):
    """
    Systematically vary input pattern feature values and collate the responses.

    Each sheet has a DistributionMatrix for each feature that will be
    tested.  The DistributionMatrix stores the distribution of
    activity values for each unit in the sheet for that feature.  For
    instance, if the features to be tested are orientation and phase,
    we will create a DistributionMatrix for orientation and a
    DistributionMatrix for phase for each sheet.  The orientation and
    phase of the input are then systematically varied (when
    measure_responses is called), and the responses of each unit
    to each pattern are collected into the DistributionMatrix.

    The resulting data can then be used to plot feature maps and
    tuning curves, or for similar types of feature-based analyses.
    """
    
    # CEB: we might want to measure the map on a sheet due
    # to a specific projection, rather than measure the map due
    # to all projections.
    
    repetitions = param.Integer(default=1,bounds=(1,None),doc="""
        How many times each stimulus will be presented.

        Each stimulus is specified by a particular feature
        combination, and need only be presented once if the network
        has no other source of variability.  If results differ for
        each presentation of an identical stimulus (e.g. due to
        intrinsic noise), then this parameter can be increased
        so that results will be an average over the specified
        number of repetitions.""")
    
    _fullmatrix = {}

    def __init__(self,features,**params):
        super(FeatureResponses,self).__init__(**params)
        self.initialize_featureresponses(features)
        self.pre_analysis_session_hooks.append(save_input_generators)
        self.post_analysis_session_hooks.append(restore_input_generators)
        
    def initialize_featureresponses(self,features):
        """Create an empty DistributionMatrix for each feature and each sheet."""
        self._featureresponses = {}
        self._activities = {}
        FeatureResponses._fullmatrix = {}
        for sheet in self.sheets_to_measure():
            self._featureresponses[sheet] = {}
            self._activities[sheet]=zeros(sheet.shape)
            for f in features:
                # CEBERRORALERT: line below is missing at least
                # "keep_peak=f.keep_peak". Couldn't these things be
                # passed around in a less fragile way?
                self._featureresponses[sheet][f.name]=DistributionMatrix(sheet.shape,axis_range=f.range,cyclic=f.cyclic) 
            FeatureResponses._fullmatrix[sheet] = FullMatrix(sheet.shape,features)

    def sheets_to_measure(self):
        """Return a list of the Sheets in the current simulation for which to collect responses."""
        return  [x for x in topo.sim.objects(Sheet).values()
                 if hasattr(x,'measure_maps') and x.measure_maps]
        
    def measure_responses(self,pattern_presenter,param_dict,features,display):
        """Present the given input patterns and collate the responses."""
        
        # Run hooks before the analysis session
        for f in self.pre_analysis_session_hooks: f()

        self.param_dict=param_dict
        self.pattern_presenter = pattern_presenter
        
        features_to_permute = [f for f in features if f.compute_fn is None]
        self.features_to_compute = [f for f in features if f.compute_fn is not None]

        self.feature_names=[f.name for f in features_to_permute]
        values_lists=[f.values for f in features_to_permute]
        self.permutations = cross_product(values_lists)
        values_description=' * '.join(["%d %s" % (len(f.values),f.name) for f in features_to_permute])
        
        self.refresh_act_wins=False
        if display:
            if hasattr(topo,'guimain'):
                self.refresh_act_wins=True
            else:
                self.warning("No GUI available for display.")

        # CEBALERT: when there are multiple sheets, this can make it seem
        # like topographica's stuck in a loop (because the counter goes
        # to 100% lots of times...e.g. hierarchical's orientation tuning fullfield.)

        timer = copy.copy(topo.sim.timer)
        timer.func = self.present_permutation

        if hasattr(topo,'guimain'):
            topo.guimain.open_progress_window(timer)
        else:
            self.verbose("Presenting %d test patterns (%s)." % (len(self.permutations),values_description))

        timer.call_fixed_num_times(self.permutations)
        
        # Run hooks after the analysis session
        for f in self.post_analysis_session_hooks: f()

    def present_permutation(self,permutation):
        """Present a pattern with the specified set of feature values."""
        for sheet in self.sheets_to_measure():
            self._activities[sheet]*=0

        # Calculate complete set of settings
        permuted_settings = zip(self.feature_names, permutation)
        complete_settings = permuted_settings + \
            [(f.name,f.compute_fn(permuted_settings)) for f in self.features_to_compute]


        for i in xrange(0,self.repetitions):
            topo.sim.state_push()

            # Run hooks before and after pattern presentation.
            # Could use complete_settings here, to avoid some
            # PatternPresenter special cases, but that might cause
            # conflicts with the existing PatternPresenter code.
            for f in self.pre_presentation_hooks: f()
            #valstring = " ".join(["%s=%s" % (n,v) for n,v in complete_settings])
            #self.message("Presenting pattern %s" % valstring)
            self.pattern_presenter(dict(permuted_settings),self.param_dict)
            for f in self.post_presentation_hooks: f()
    
            if self.refresh_act_wins:topo.guimain.refresh_activity_windows()
            for sheet in self.sheets_to_measure():
                self._activities[sheet]+=sheet.activity
            topo.sim.state_pop()

        for sheet in self.sheets_to_measure():
            self._activities[sheet]=self._activities[sheet] / self.repetitions
        
        self._update(complete_settings)
         
    def _update(self,current_values):
        # Update each DistributionMatrix with (activity,bin)
        for sheet in self.sheets_to_measure():
            for feature,value in current_values:
                self._featureresponses[sheet][feature].update(self._activities[sheet], value)
            FeatureResponses._fullmatrix[sheet].update(self._activities[sheet],current_values)



class ReverseCorrelation(FeatureResponses):
    """
    Calculate the receptive fields for all neurons using reverse correlation.
    """
    # CB: Can't we have a better class hierarchy?

    input_sheet = param.Parameter(default=None)

    # JABALERT: Should _featureresponses be renamed here?; It's a different
    # data structure using different indexing (r,c instead of feature).
    def initialize_featureresponses(self,features): # CB: doesn't need features!

        self._featureresponses = {}
        assert hasattr(self.input_sheet,'shape')
        
        # surely there's a way to get an array of 0s for each element without
        # looping? (probably had same question for distributionmatrix).
        for sheet in self.sheets_to_measure():
            self._featureresponses[sheet]= numpy.ones(sheet.activity.shape,dtype=object) 
            rows,cols = sheet.activity.shape
            for r in range(rows):
                for c in range(cols):
                    self._featureresponses[sheet][r,c] = numpy.zeros(self.input_sheet.shape) # need to specify dtype?
        

    def collect_feature_responses(self,pattern_presenter,param_dict,display,feature_values):
        self.measure_responses(pattern_presenter,param_dict,feature_values,display)

        for sheet in self.sheets_to_measure():
            rows,cols = sheet.activity.shape
            input_bounds = self.input_sheet.bounds
            input_sheet_views = self.input_sheet.sheet_views

            for ii in range(rows):
                for jj in range(cols):
                    view = SheetView((self._featureresponses[sheet][ii,jj],input_bounds),
                                     sheet.name,sheet.precedence,topo.sim.time(),sheet.row_precedence)
                    x,y = sheet.matrixidx2sheet(ii,jj)
                    key = ('RFs',sheet.name,x,y)
                    input_sheet_views[key]=view

                    
    

    def measure_responses(self,pattern_presenter,param_dict,features,display):
        """Present the given input patterns and collate the responses."""

        # Since input_sheet's not fixed, we have to call this. Means that there are
        # normally duplicate calls (e.g. gets called by __init__ and then gets called
        # here for no reason except maybe the input_sheet got changed). Would be better
        # to have the input_sheet fixed.
        self.initialize_featureresponses(features) 

        super(ReverseCorrelation,self).measure_responses(pattern_presenter,param_dict,
                                                         features,display)
                                                         
    def present_permutation(self,permutation):
        """Present a pattern with the specified set of feature values."""

        # Calculate complete set of settings
        permuted_settings = zip(self.feature_names, permutation)
        complete_settings = permuted_settings + \
            [(f.name,f.compute_fn(permuted_settings)) for f in self.features_to_compute]

        topo.sim.state_push()

        # Run hooks before and after pattern presentation.
        # Could use complete_settings here, to avoid some
        # PatternPresenter special cases, but that might cause
        # conflicts with the existing PatternPresenter code.
        for f in self.pre_presentation_hooks: f()
        #valstring = " ".join(["%s=%s" % (n,v) for n,v in complete_settings])
        #self.message("Presenting pattern %s" % valstring)
        self.pattern_presenter(dict(permuted_settings),self.param_dict)
        for f in self.post_presentation_hooks: f()
    
        if self.refresh_act_wins:topo.guimain.refresh_activity_windows()

        self._update(complete_settings)
         
        topo.sim.state_pop()

    # Ignores current_values; they simply provide distinct patterns on the retina
    def _update(self,current_values):
        for sheet in self.sheets_to_measure():
            rows,cols = sheet.activity.shape
            for ii in range(rows): 
                for jj in range(cols):
                    self._featureresponses[sheet][ii,jj]+=sheet.activity[ii,jj]*self.input_sheet.activity
                    

class FeatureMaps(FeatureResponses):
    """
    Measures and collects the responses to a set of features for calculating feature maps.

    For each feature and each sheet, the results are stored as a
    preference matrix and selectivity matrix in the sheet's
    sheet_views; these can then be plotted as preference
    or selectivity maps.
    """

    preference_fn = param.ClassSelector( DistributionStatisticFn, default=DSF_WeightedAverage(),
            doc=""" Function that will be used to analyze the distributions of unit responses.
            Note that this default is orverriden by specific functions for features, if
            specified in the Feature objects""" )


    selectivity_multiplier = param.Number(default=17.0, doc="""
            Scaling of the feature selectivity values applied in all feature dimensions.
            The multiplier sets the output scaling.""")
    
    
    # CBENHANCEMENT: could allow full control over the generated names
    # using a format parameter. The default would be
    # ${prefix}${feature}${type} (where type is Preference or
    # Selectivity)
    sheet_views_prefix = param.String(default="",doc="""
            Prefix to add to the name under which results are stored in sheet_views.""")

    def __init__(self,features,**params):
        super(FeatureMaps,self).__init__(features,**params)
        self.features=features
        
    def collect_feature_responses(self,pattern_presenter,param_dict,display):
        """
        Present the given input patterns and collate the responses.

        Responses are statistics on the distributions of measure for every unit,
        extracted by functions that are subclasses of DistributionStatisticFn, and could
        be specified in each feature with the preference_fn parameter, otherwise the
        default in self.preference_fn is used.
        """
        self.measure_responses(pattern_presenter,param_dict,self.features,display)    

        for sheet in self.sheets_to_measure():
            bounding_box    = sheet.bounds
            sn              = sheet.name
            sp              = sheet.precedence
            sr              = sheet.row_precedence
            
            for feature in self._featureresponses[sheet].keys():
            ### JCHACKALERT! This is temporary to avoid the positionpref plot to shrink
            ### Nevertheless we should think more about this (see alert in bitmap.py)
            ### When passing a sheet_view that is not cropped to 1 in the parameter hue of hsv_to_rgb
            ### it does not work... The normalization seems to be necessary in this case.
            ### I guess it is always cyclic value that we will color with hue in an hsv plot
            ### but still we should catch the error.
            ### Also, what happens in case of negative values?
                fp                  = filter( lambda f: f.name==feature, self.features )[ 0 ]
                ar                  = self._featureresponses[sheet][feature].distribution_matrix[0,0].axis_range
                cyclic              = fp.cyclic
                cyclic_range        = ar    if cyclic   else 1.0
                preference_fn       = fp.preference_fn if fp.preference_fn is not None else self.preference_fn
                if self.selectivity_multiplier is not None:
                    preference_fn.selectivity_scale = (preference_fn.selectivity_scale[0], self.selectivity_multiplier)
                fr                  = self._featureresponses[sheet][feature]
                response            = fr.apply_DSF( preference_fn )
                base_name           = self.sheet_views_prefix + feature.capitalize()

                for k, maps in response.items():
                    t           = topo.sim.time()
                    for map_name, map_view in maps.items():
                        name                        = base_name + k + map_name.capitalize()
                        # JABALERT: Is this safe and general enough?
                        if map_name == 'selectivity':
                            cyclic          = False
                            cyclic_range    = None
                        view                        = SheetView( (map_view, bounding_box), sn, sp, t, sr )
                        view.cyclic                 = cyclic
                        view.cyclic_range           = cyclic_range
                        sheet.sheet_views[ name ]   = view

                
class FeatureCurves(FeatureResponses):
    """
    Measures and collects the responses to a set of features, for calculating tuning and similar curves.    

    These curves represent the response of a Sheet to patterns that
    are controlled by a set of features.  This class can collect data
    for multiple curves, each with the same x axis.  The x axis
    represents the main feature value that is being varied, such as
    orientation.  Other feature values can also be varied, such as
    contrast, which will result in multiple curves (one per unique
    combination of other feature values).

    The sheet responses used to construct the curves will be stored in
    a dictionary curve_dict kept in the Sheet of interest.  A
    particular set of patterns is then constructed using a
    user-specified PatternPresenter by adding the parameters
    determining the curve (curve_param_dict) to a static list of
    parameters (param_dict), and then varying the specified set of
    features.  The results can be accessed in the curve_dict,
    indexed by the curve_label and feature value.
    """
    post_collect_responses_hook = param.HookList(default=[],instantiate=False,doc="""
        List of callable objects to be run at the end of collect_feature_responses function.
        The functions should accept three parameters: FullMatrix, curve label, sheet""")

    def __init__(self,features,sheet,x_axis):
        super(FeatureCurves, self).__init__(features)
        self.sheet=sheet
        self.x_axis=x_axis
        if hasattr(sheet, "curve_dict")==False:
            sheet.curve_dict={}
        sheet.curve_dict[x_axis]={}

    def sheets_to_measure(self):
        return topo.sim.objects(CFSheet).values()

    def collect_feature_responses(self,features,pattern_presenter,param_dict,curve_label,display):
        self.initialize_featureresponses(features)
        rows,cols=self.sheet.shape
        bounding_box = self.sheet.bounds
        self.measure_responses(pattern_presenter,param_dict,features,display)
        self.sheet.curve_dict[self.x_axis][curve_label]={}
        for key in self._featureresponses[self.sheet][self.x_axis].distribution_matrix[0,0]._data.iterkeys():
            y_axis_values = zeros(self.sheet.shape,activity_type)
            for i in range(rows):
                for j in range(cols):
                    y_axis_values[i,j] = self._featureresponses[self.sheet][self.x_axis].distribution_matrix[i,j].get_value(key)
            Response = SheetView((y_axis_values,bounding_box), self.sheet.name , self.sheet.precedence, topo.sim.time(),self.sheet.row_precedence)
            self.sheet.curve_dict[self.x_axis][curve_label].update({key:Response})
        for f in self.post_collect_responses_hook: f(self._fullmatrix[self.sheet],curve_label,self.sheet)

###############################################################################
###############################################################################

# Define user-level commands and helper classes for calling the above


class Feature( param.Parameterized ):
    """
    Specifies several parameters required for generating a map of one input feature

    """

    name            = param.String( default="", doc="Name of the feature to test" )

    cyclic          = param.Boolean( default=False, doc="""
            Whether the range of this feature is cyclic (wraps around at the high end)""" )

    compute_fn      = param.Callable( default=None, doc="""
            If non-None, a function that when given a list of other parameter values,
            computes and returns the value for this feature""" )

    preference_fn   = param.ClassSelector( DistributionStatisticFn, default=DSF_WeightedAverage(),
            doc="""Function that will be used to analyze the distributions of unit response
            to this feature""")

    range           = param.NumericTuple( default=(0,0), doc="""
            lower and upper values for a feature, used to build a list of values,
            together with the step parameter""" )

    step            = param.Number( default=0.0, doc="""
            increment used to build a list of values for this feature, together with
            the range parameter""" )

    offset          = param.Number( default=0.0, doc="offset to add to the values for this feature" )

    values          = param.List(default=[], doc="""
            explicit list of values for this feature, used in alternative to the range
            and step parameters""" )


    def __init__(self, **params ):
        """
        Users can provide either a range and a step size, or a list of values.
        If a list of values is supplied, the range can be omitted unless the
        default of the min and max in the list of values is not appropriate.

        If non-None, the compute_fn should be a function that when given a list 
        of other parameter values, computes and returns the value for this feature.

        If supplied, the offset is added to the given or computed values to allow
        the starting value to be specified.

        """

        super( Feature, self ).__init__( **params )
                 
        if len( self.values ):
            self.values = self.values if self.offset == 0 else [v+self.offset for v in self.values]
            if self.range == ( 0, 0 ):
                self.range = ( min(self.values), max(self.values) )
        else:
            if self.range == ( 0, 0 ):
                raise ValueError('The range or values must be specified.')
            low_bound,up_bound = self.range
            self.values = frange( low_bound, up_bound, self.step, not self.cyclic )
            self.values = self.values if self.offset == 0 else \
                    [(v + self.offset) % (up_bound - low_bound) if self.cyclic else (v + self.offset)
                    for v in self.values]


class PatternPresenter(param.Parameterized):
    """
    Function object for presenting PatternGenerator-created patterns.

    This class helps coordinate a set of patterns to be presented to a
    set of GeneratorSheets.  It provides a standardized way of
    generating a set of linked patterns for testing or analysis, such
    as when measuring preference maps or presenting test patterns.
    Subclasses can provide additional mechanisms for doing this in
    different ways.
    """
    
    # JABALERT: Needs documenting, and probably also a clearer name
    contrast_parameter = param.Parameter('michelson_contrast')

    # JABALERT: Needs documenting; apparently only for retinotopy?
    divisions = param.Parameter()

    apply_output_fns = param.Boolean(default=True, doc="""
        When presenting a pattern, whether to apply each sheet's
        output function.  If False, for many networks the response
        will be linear, which requires fewer test patterns to measure
        a meaningful response, but it may not correspond to the actual
        preferences of each neuron under other conditions.  If True,
        callers will need to ensure that the input patterns are in a
        suitable range to drive the neurons to generate meaningful
        output, because e.g. a threshold-based output function might
        result in no activity for inputs that are too weak..""")

    duration = param.Number(default=1.0,doc="""
        Amount of simulation time for which to present each test pattern.
        By convention, most Topographica example files are designed to
        have a suitable activity pattern computed by the
        default time, but the duration will need to be changed for
        other models that do not follow that convention or if a
        linear response is desired.""")

    # CEBALERT: generator_sheets=[] is probably a surprising way of
    # actually getting all the generator sheets.
    generator_sheets = param.List(default=[], doc="""
        The set of GeneratorSheets onto which patterns will be drawn.
        
        By default (i.e. for an empty list), all GeneratorSheets in
        the simulation will be used.
        """)

    def __init__(self,pattern_generator,**params):
        """
        pattern_generator is the PatternGenerator that will be drawn
        on the generator_sheets (the parameters of the
        pattern_generator are specified during calls.
        """
        super(PatternPresenter,self).__init__(**params)
        self.gen = pattern_generator # Why not a Parameter?

        
    def __call__(self,features_values,param_dict):
        for param,value in param_dict.iteritems():
            # CEBALERT: why not setattr(self.gen,param,value)
            #if ('_'+param+'_param_value') not in self.gen.__dict__:
            self.gen.__setattr__(param,value)                                
                              
        for feature,value in features_values.iteritems():
            self.gen.__setattr__(feature,value)
        
        all_input_sheet_names = topo.sim.objects(GeneratorSheet).keys()

        if len(self.generator_sheets)>0:
            input_sheet_names = [sheet.name for sheet in self.generator_sheets]
        else:
            input_sheet_names = all_input_sheet_names
        
        # Copy the given generator once for every GeneratorSheet
        inputs = dict.fromkeys(input_sheet_names)
        for k in inputs.keys():
            inputs[k]=copy.deepcopy(self.gen)
        
        ### JABALERT: Should replace these special cases with general
        ### support for having meta-parameters controlling the
        ### generation of different patterns for each GeneratorSheet.
        ### For instance, we will also need to support xdisparity and
        ### ydisparity, plus movement of patterns between two eyes, colors,
        ### etc.  At the very least, it should be simple to control
        ### differences in single parameters easily.  In addition,
        ### these meta-parameters should show up as parameters for
        ### this object, augmenting the parameters for each individual
        ### pattern, e.g. in the Test Pattern window.  In this way we
        ### should be able to provide general support for manipulating
        ### both pattern parameters and parameters controlling
        ### interaction between or differences between patterns.
        
        if 'direction' in features_values:
            import __main__
            if '_new_motion_model' in __main__.__dict__ and __main__.__dict__['_new_motion_model']:
            #### new motion model ####
                from topo.pattern import Translator
                for name in inputs:
                    inputs[name] = Translator(generator=inputs[name],
                                              direction=features_values['direction'],
                                              speed=features_values['speed'],
                                              reset_period=self.duration) 
            ##########################
            else:
            #### old motion model ####
                orientation = features_values['direction']+pi/2
                from topo.pattern import Sweeper            
                for name in inputs.keys():
                    speed=features_values['speed']
                    try:
                        step=int(name[-1])
                    except:
                        if not hasattr(self,'direction_warned'):
                            self.warning('Assuming step is zero; no input lag number specified at the end of the input sheet name.')
                            self.direction_warned=True
                        step=0
                    speed=features_values['speed']
                    inputs[name] = Sweeper(generator=inputs[name],step=step,speed=speed)
                    setattr(inputs[name],'orientation',orientation)
            ##########################
        
        if features_values.has_key('hue'):

            # could be three retinas (R, G, and B) or a single RGB
            # retina for the color dimension; if every retina has
            # 'Red' or 'Green' or 'Blue' in its name, then three
            # retinas for color are assumed

            rgb_retina = False 
            for name in input_sheet_names:
                if not ('Red' in name or 'Green' in name or 'Blue' in name):
                    rgb_retina=True

            if not rgb_retina:
                for name in inputs.keys():
                    r,g,b=hsv_to_rgb(features_values['hue'],1.0,1.0)
                    if (name.count('Red')):
                        inputs[name].scale=r
                    elif (name.count('Green')):
                        inputs[name].scale=g
                    elif (name.count('Blue')):
                        inputs[name].scale=b
                    else: 
                        if not hasattr(self,'hue_warned'):
                            self.warning('Unable to measure hue preference, because hue is defined only when there are different input sheets with names with Red, Green or Blue substrings.')
                            self.hue_warned=True
            else:
                from contrib import rgbimages
                r,g,b=hsv_to_rgb(features_values['hue'],1.0,1.0)
                for name in inputs.keys():
                    inputs[name] = rgbimages.ExtendToRGB(generator=inputs[name],
                                                         relative_channel_strengths=[r,g,b])
                # CEBALERT: should warn as above if not a color network
        
        #JL: This is only used for retinotopy measurement in jude laws contrib/jsldefs.py
        #Also needs cleaned up
        if features_values.has_key('retinotopy'):
            #Calculates coordinates of the center of each patch to be presented 
            coordinate_x=[]
            coordinate_y=[]
            coordinates=[]
            for name,i in zip(inputs.keys(),range(len(input_sheet_names))):
                l,b,r,t = topo.sim[name].nominal_bounds.lbrt()
                x_div=float(r-l)/(self.divisions*2)
                y_div=float(t-b)/(self.divisions*2)
                for i in range(self.divisions):
                    if not bool(self.divisions%2):
                        if bool(i%2):
                            coordinate_x.append(i*x_div)
                            coordinate_y.append(i*y_div)
                            coordinate_x.append(i*-x_div)
                            coordinate_y.append(i*-y_div)
                    else:
                        if not bool(i%2):
                            coordinate_x.append(i*x_div)
                            coordinate_y.append(i*y_div)
                            coordinate_x.append(i*-x_div)
                            coordinate_y.append(i*-y_div)
                for x in coordinate_x:
                    for y in coordinate_y:
                        coordinates.append((x,y))
               
                x_coord=coordinates[features_values['retinotopy']][0]
                y_coord=coordinates[features_values['retinotopy']][1]
                inputs[name].x = x_coord
                inputs[name].y = y_coord
          
        if features_values.has_key('retx'):
            for name,i in zip(inputs.keys(),range(len(input_sheet_names))):
                inputs[name].x = features_values['retx']
                inputs[name].y = features_values['rety']           
                
        if features_values.has_key("phasedisparity"):
            temp_phase1=features_values['phase']-features_values['phasedisparity']/2.0
            temp_phase2=features_values['phase']+features_values['phasedisparity']/2.0
            for name in inputs.keys():
                if (name.count('Right')):
                    inputs[name].phase=wrap(0,2*pi,temp_phase1)
                elif (name.count('Left')):
                    inputs[name].phase=wrap(0,2*pi,temp_phase2)
                else:
                    if not hasattr(self,'disparity_warned'):
                        self.warning('Unable to measure disparity preference, because disparity is defined only when there are inputs for Right and Left retinas.')
                        self.disparity_warned=True
                
        ## Not yet used; example only
        #if features_values.has_key("xdisparity"):
        #    if len(input_sheet_names)!=2:
        #        self.warning('Disparity is defined only when there are exactly two patterns')
        #    else:
        #        inputs[input_sheet_names[0]].x=inputs[input_sheet_names[0]].x - inputs[input_sheet_names[0]].xdisparity/2.0
        #        inputs[input_sheet_names[1]].x=inputs[input_sheet_names[1]].x + inputs[input_sheet_names[1]].xdisparity/2.0
        #
        #        inputs={}
        #        inputs[input_sheet_names[0]]=inputs[input_sheet_names[0]]
        #        inputs[input_sheet_names[1]]=inputs[input_sheet_names[1]]
        #
        #if features_values.has_key("ydisparity"):
        #    if len(input_sheet_names)!=2:
        #        self.warning('Disparity is defined only when there are exactly two patterns')
        #    else:
        #        inputs[input_sheet_names[0]].y=inputs[input_sheet_names[0]].y - inputs[input_sheet_names[0]].ydisparity/2.0
        #        inputs[input_sheet_names[1]].y=inputs[input_sheet_names[1]].y + inputs[input_sheet_names[1]].ydisparity/2.0
        #
        #        inputs={}
        #        inputs[input_sheet_names[0]]=inputs[input_sheet_names[0]]
        #        inputs[input_sheet_names[1]]=inputs[input_sheet_names[1]]

        if features_values.has_key("ocular"):
            for name in inputs.keys():
                if (name.count('Right')):
                    inputs[name].scale=2*features_values['ocular']
                elif (name.count('Left')):
                    inputs[name].scale=2.0-2*features_values['ocular']
                else:
                    self.warning('Skipping input region %s; Ocularity is defined only for Left and Right retinas.' %
                                 name)

        if features_values.has_key("contrastcenter")or param_dict.has_key("contrastcenter"):
            if self.contrast_parameter=='michelson_contrast':
                for g in inputs.itervalues():
                    g.offsetcenter=0.5
                    g.scalecenter=2*g.offsetcenter*g.contrastcenter/100.0
        
            elif self.contrast_parameter=='weber_contrast':
                # Weber_contrast is currently only well defined for
                # the special case where the background offset is equal
                # to the target offset in the pattern type
                # SineGrating(mask_shape=Disk())
                for g in inputs.itervalues():
                    g.offsetcenter=0.5   #In this case this is the offset of both the background and the sine grating
                    g.scalecenter=2*g.offsetcenter*g.contrastcenter/100.0
            
            elif self.contrast_parameter=='scale':
                for g in inputs.itervalues():
                    g.offsetcenter=0.0
                    g.scalecenter=g.contrastcenter

        if features_values.has_key("contrastsurround")or param_dict.has_key("contrastsurround"):
            if self.contrast_parameter=='michelson_contrast':
                for g in inputs.itervalues():
                    g.offsetsurround=0.5
                    g.scalesurround=2*g.offsetsurround*g.contrastsurround/100.0
        
            elif self.contrast_parameter=='weber_contrast':
                # Weber_contrast is currently only well defined for
                # the special case where the background offset is equal
                # to the target offset in the pattern type
                # SineGrating(mask_shape=Disk())
                for g in inputs.itervalues():
                    g.offsetsurround=0.5   #In this case this is the offset of both the background and the sine grating
                    g.scalesurround=2*g.offsetsurround*g.contrastsurround/100.0
                
            elif self.contrast_parameter=='scale':
                for g in inputs.itervalues():
                    g.offsetsurround=0.0
                    g.scalesurround=g.contrastsurround
        
        if features_values.has_key("contrast") or param_dict.has_key("contrast"):
            if self.contrast_parameter=='michelson_contrast':
                for g in inputs.itervalues():
                    g.offset=0.5
                    g.scale=2*g.offset*g.contrast/100.0
        
            elif self.contrast_parameter=='weber_contrast':
                # Weber_contrast is currently only well defined for
                # the special case where the background offset is equal
                # to the target offset in the pattern type
                # SineGrating(mask_shape=Disk())
                for g in inputs.itervalues():
                    g.offset=0.5   #In this case this is the offset of both the background and the sine grating
                    g.scale=2*g.offset*g.contrast/100.0
                
            elif self.contrast_parameter=='scale':
                for g in inputs.itervalues():
                    g.offset=0.0
                    g.scale=g.contrast

        # blank patterns for unused generator sheets
        for sheet_name in set(all_input_sheet_names).difference(set(input_sheet_names)):
            inputs[sheet_name]=pattern.Constant(scale=0)
                
        pattern_present(inputs, self.duration, plastic=False,
                     apply_output_fns=self.apply_output_fns)


class Subplotting(param.Parameterized):
    """
    Convenience functions for handling subplots (such as colorized Activity plots).
    Only needed for avoiding typing, as plots can be declared with their own
    specific subplots without using these functions.
    """
    
    plotgroups_to_subplot=param.List(default=
        ["Activity", "Connection Fields", "Projection", "Projection Activity"],
        doc="List of plotgroups for which to set subplots.")

    subplotting_declared = param.Boolean(default=False,
        doc="Whether set_subplots has previously been called")
    
    _last_args = param.Parameter(default=())
    
    @staticmethod    
    def set_subplots(prefix=None,hue="",confidence="",force=True):
        """
        Define Hue and Confidence subplots for each of the plotgroups_to_subplot.  
        Typically used to make activity or weight plots show a
        preference value as the hue, and a selectivity as the
        confidence.

        The specified hue, if any, should be the name of a SheetView,
        such as OrientationPreference.  The specified confidence, if
        any, should be the name of a (usually) different SheetView,
        such as OrientationSelectivity.

        The prefix option is a shortcut making the usual case easier
        to type; it sets hue to prefix+"Preference" and confidence to
        prefix+"Selectivity".

        If force=False, subplots are changed only if no subplot is
        currently defined.  Force=False is useful for setting up
        subplots automatically when maps are measured, without
        overwriting any subplots set up specifically by the user.

        Currently works only for plotgroups that have a plot
        with the same name as the plotgroup, though this could
        be changed easily.

        Examples::
        
           Subplotting.set_subplots("Orientation")
             - Set the default subplots to OrientationPreference and OrientationSelectivity

           Subplotting.set_subplots(hue="OrientationPreference")
             - Set the default hue subplot to OrientationPreference with no selectivity

           Subplotting.set_subplots()
             - Remove subplots from all the plotgroups_to_subplot.
        """

        Subplotting._last_args=(prefix,hue,confidence,force)
        
        if Subplotting.subplotting_declared and not force:
            return

        if prefix:
            hue=prefix+"Preference"
            confidence=prefix+"Selectivity"
        
        for name in Subplotting.plotgroups_to_subplot:
            if plotgroups.has_key(name):
                pg=plotgroups[name]
                if pg.plot_templates.has_key(name):
                    pt=pg.plot_templates[name]
                    pt["Hue"]=hue
                    pt["Confidence"]=confidence
                else:
                    Subplotting().warning("No template %s defined for plotgroup %s" % (name,name))
            else:
                Subplotting().warning("No plotgroup %s defined" % name)

        Subplotting.subplotting_declared=True


    @staticmethod    
    def restore_subplots():
        args=Subplotting._last_args
        if args != (): Subplotting.set_subplots(*(Subplotting._last_args))



###############################################################################
###############################################################################
###############################################################################
#
# 20081017 JABNOTE: This implementation could be improved.  
#
# It currently requires every subclass to implement the feature_list
# method, which constructs a list of features using various parameters
# to determine how many and which values each feature should have.  It
# would be good to replace the feature_list method with a Parameter or
# set of Parameters, since it is simply a special data structure, and
# this would make more detailed control feasible for users. For
# instance, instead of something like num_orientations being used to
# construct the orientation Feature, the user could specify the
# appropriate Feature directly, so that they could e.g. supply a
# specific list of orientations instead of being limited to a fixed
# spacing.
# 
# However, when we implemented this, we ran into two problems:
#
# 1. It's difficult for users to modify an open-ended list of
#     Features.  E.g., if features is a List:
#
#      features=param.List(doc="List of Features to vary""",default=[
#          Feature(name="frequency",values=[2.4]),
#          Feature(name="orientation",range=(0.0,pi),step=pi/4,cyclic=True),
#          Feature(name="phase",range=(0.0,2*pi),step=2*pi/18,cyclic=True)])
#
#    then it it's easy to replace the entire list, but tough to
#    change just one Feature.  Of course, features could be a
#    dictionary, but that doesn't help, because when the user
#    actually calls the function, they want the arguments to
#    affect only that call, whereas looking up the item in a
#    dictionary would only make permanent changes easy, not
#    single-call changes.
#    
#    Alternatively, one could make each feature into a separate     
#    parameter, and then collect them using a naming convention like:
#
#     def feature_list(self,p):
#         fs=[]
#         for n,v in self.get_param_values():
#             if n in p: v=p[n]
#             if re.match('^[^_].*_feature$',n):
#                 fs+=[v]
#         return fs
#    
#    But that's quite hacky, and doesn't solve problem 2.
# 
# 2. Even if the users can somehow access each Feature, the same
#    problem occurs for the individual parts of each Feature.  E.g.
#    using the separate feature parameters above, Spatial Frequency
#    map measurement would require:
#
#      from topo.command.analysis import Feature
#      from math import pi
#      pre_plot_hooks=[measure_or_pref.instance( \
#         frequency_feature=Feature(name="frequency",values=frange(1.0,6.0,0.2)), \
#         phase_feature=Feature(name="phase",range=(0.0,2*pi),step=2*pi/15,cyclic=True), \
#         orientation_feature=Feature(name="orientation",range=(0.0,pi),step=pi/4,cyclic=True)])
#   
#    rather than the current, much more easily controllable implementation:
#   
#      pre_plot_hooks=[measure_or_pref.instance(frequencies=frange(1.0,6.0,0.2),\
#         num_phase=15,num_orientation=4)]
#
#    I.e., to change anything about a Feature, one has to supply an
#    entirely new Feature, because otherwise the original Feature
#    would be changed for all future calls.  Perhaps there's some way
#    around this by copying objects automatically at the right time,
#    but if so it's not obvious.  Meanwhile, the current
#    implementation is reasonably clean and easy to use, if not as
#    flexible as it could be.



class MeasureResponseCommand(ParameterizedFunction):
    """Parameterized command for presenting input patterns and measuring responses."""
      
    scale = param.Number(default=1.0,softbounds=(0.0,2.0),doc="""
        Multiplicative strength of input pattern.""")
    
    offset = param.Number(default=0.0,softbounds=(-1.0,1.0),doc="""
        Additive offset to input pattern.""")

    display = param.Boolean(default=False,doc="""
        Whether to update a GUI display (if any) during the map measurement.""")

    weighted_average= param.Boolean(default=True, doc="""
        Whether to compute results using a weighted average, or just
        discrete values.  A weighted average can give more precise
        results, without being limited to a set of discrete values,
        but the results can have systematic biases due to the
        averaging, especially for non-cyclic parameters.""")

    pattern_presenter = param.Callable(default=None,instantiate=True,doc="""
        Callable object that will present a parameter-controlled pattern to a
        set of Sheets.  Needs to be supplied by a subclass or in the call.
        The attributes duration and apply_output_fns (if non-None) will
        be set on this object, and it should respect those if possible.""")
    
    static_parameters = param.List(class_=str,default=["scale","offset"],doc="""
        List of names of parameters of this class to pass to the
        pattern_presenter as static parameters, i.e. values that
        will be fixed to a single value during measurement.""")

    subplot = param.String("",doc="""Name of map to register as a subplot, if any.""")

    apply_output_fns = param.Boolean(default=None, doc="""
        If non-None, pattern_presenter.apply_output_fns will be
        set to this value.  Provides a simple way to set
        this commonly changed option of PatternPresenter.""")

    duration = param.Number(default=None,doc="""
        If non-None, pattern_presenter.duration will be
        set to this value.  Provides a simple way to set
        this commonly changed option of PatternPresenter.""")

    sheet_views_prefix = param.String(default="",doc="""
        Optional prefix to add to the name under which results are
        stored in sheet_views. Can be used e.g. to distinguish maps as
        originating from a particular GeneratorSheet.""")

    generator_sheets = param.List(default=[],doc="""
        pattern_presenter.generator_sheets will be set to this value.
        The default value of [] results in all GeneratorSheets being
        used.""")

    preference_fn = param.ClassSelector( DistributionStatisticFn, default=DSF_MaxValue(),
            doc="""Function that will be used to analyze the distributions of unit responses.""" )

    __abstract = True


    def __call__(self,**params):
        """Measure the response to the specified pattern and store the data in each sheet."""

        p=ParamOverrides(self,params)
        x=FeatureMaps(self._feature_list(p),name="FeatureMaps_for_"+self.name,
                      sheet_views_prefix=p.sheet_views_prefix)
        static_params = dict([(s,p[s]) for s in p.static_parameters])
        if p.duration is not None:
            p.pattern_presenter.duration=p.duration
        if p.apply_output_fns is not None:
            p.pattern_presenter.apply_output_fns=p.apply_output_fns
        p.pattern_presenter.generator_sheets=p.generator_sheets

        x.collect_feature_responses(p.pattern_presenter,static_params,
                                    p.display)

        if p.subplot != "":
            Subplotting.set_subplots(p.subplot,force=True)

        return x._fullmatrix


    def _feature_list(self,p):
        """Return the list of features to vary; must be implemented by each subclass."""
        raise NotImplementedError

    

class SinusoidalMeasureResponseCommand(MeasureResponseCommand):
    """Parameterized command for presenting sine gratings and measuring responses."""
    
    pattern_presenter = param.Callable(instantiate=True,
        default=PatternPresenter(pattern_generator=SineGrating()),doc="""
        Callable object that will present a parameter-controlled pattern to a
        set of Sheets.  By default, uses a SineGrating presented for a short
        duration.  By convention, most Topographica example files
        are designed to have a suitable activity pattern computed by
        that time, but the duration will need to be changed for other
        models that do not follow that convention.""")

    frequencies = param.List(class_=float,default=[2.4],doc="Sine grating frequencies to test.")
    
    num_phase = param.Integer(default=18,bounds=(1,None),softbounds=(1,48),
                              doc="Number of phases to test.")

    num_orientation = param.Integer(default=4,bounds=(1,None),softbounds=(1,24),
                                    doc="Number of orientations to test.")

    scale = param.Number(default=0.3)

    preference_fn = param.ClassSelector( DistributionStatisticFn, default=DSF_WeightedAverage(),
            doc="""Function that will be used to analyze the distributions of unit responses.""" )

    __abstract = True
    


class PositionMeasurementCommand(MeasureResponseCommand):
    """Parameterized command for measuring topographic position."""

    divisions=param.Integer(default=6,bounds=(1,None),doc="""
        The number of different positions to measure in X and in Y.""")
    
    x_range=param.NumericTuple((-0.5,0.5),doc="""
        The range of X values to test.""")

    y_range=param.NumericTuple((-0.5,0.5),doc="""
        The range of Y values to test.""")

    size=param.Number(default=0.5,bounds=(0,None),doc="""
        The size of the pattern to present.""")
    
    pattern_presenter = param.Callable(
        default=PatternPresenter(Gaussian(aspect_ratio=1.0)),doc="""
        Callable object that will present a parameter-controlled
        pattern to a set of Sheets.  For measuring position, the
        pattern_presenter should be spatially localized, yet also able
        to activate the appropriate neurons reliably.""")

    static_parameters = param.List(default=["scale","offset","size"])
  
    __abstract = True

        


class SingleInputResponseCommand(MeasureResponseCommand):
    """
    A callable Parameterized command for measuring the response to input on a specified Sheet.

    Note that at present the input is actually presented to all input sheets; the
    specified Sheet is simply used to determine various parameters.  In the future,
    it may be modified to draw the pattern on one input sheet only.
    """
    # CBERRORALERT: Need to alter PatternPresenter to accept an input sheet,
    # to allow it to be presented on only one sheet.
    
    input_sheet = param.ObjectSelector(
        default=None,doc="""
        Name of the sheet where input should be drawn.""")

    scale = param.Number(default=30.0)

    offset = param.Number(default=0.5)

    # JABALERT: Presumably the size is overridden in the call, right?
    pattern_presenter = param.Callable(
        default=PatternPresenter(RawRectangle(size=0.1,aspect_ratio=1.0)))

    static_parameters = param.List(default=["scale","offset","size"])

    weighted_average = None # Disabled unused parameter

    __abstract = True



class FeatureCurveCommand(SinusoidalMeasureResponseCommand):
    """A callable Parameterized command for measuring tuning curves."""

    num_orientation = param.Integer(default=12)

    sheet = param.ObjectSelector(
        default=None,doc="""
        Name of the sheet to use in measurements.""")

    units = param.String(default='%',doc="""
        Units for labeling the curve_parameters in figure legends.
        The default is %, for use with contrast, but could be any 
        units (or the empty string).""")

    # Make constant in subclasses?
    x_axis = param.String(default='orientation',doc="""
        Parameter to use for the x axis of tuning curves.""")

    static_parameters = param.List(default=[])

    # JABALERT: Might want to accept a list of values for a given
    # parameter to make the simple case easier; then maybe could do
    # the crossproduct of them?
    curve_parameters=param.Parameter([{"contrast":30},{"contrast":60},{"contrast":80},{"contrast":90}],doc="""
        List of parameter values for which to measure a curve.""")
    
    __abstract = True

    def __call__(self,**params):
        """Measure the response to the specified pattern and store the data in each sheet."""
        p=ParamOverrides(self,params)
        self._compute_curves(p,p.sheet)


    def _compute_curves(self,p,sheet,val_format='%s'):
        """
        Compute a set of curves for the specified sheet, using the
        specified val_format to print a label for each value of a
        curve_parameter.
        """

        x=FeatureCurves(self._feature_list(p),sheet=sheet,x_axis=self.x_axis)
        for curve in p.curve_parameters:
            static_params = dict([(s,p[s]) for s in p.static_parameters])
            static_params.update(curve)
            curve_label="; ".join([('%s = '+val_format+'%s') % (n.capitalize(),v,p.units) for n,v in curve.items()])
            # JABALERT: Why is the feature list duplicated here?
            x.collect_feature_responses(self._feature_list(p),p.pattern_presenter,static_params,curve_label,p.display)



    def _feature_list(self,p):
        return [Feature(name="phase",range=(0.0,2*pi),step=2*pi/p.num_phase,cyclic=True),
                Feature(name="orientation",range=(0,pi),step=pi/p.num_orientation,cyclic=True),
                Feature(name="frequency",values=p.frequencies)]
        

    def _sheetview_unit(self,sheet,sheet_coord,map_name,default=0.0):
        """Look up and return the value of a SheetView for a specified unit."""
        matrix_coords = sheet.sheet2matrixidx(*sheet_coord)

        if(map_name in sheet.sheet_views):
            pref = sheet.sheet_views[map_name].view()[0]
            val = pref[matrix_coords]
        else:
            self.warning(("%s should be measured before plotting this tuning curve -- " + 
                          "using default value of %s for %s unit (%d,%d).") % \
                         (map_name,default,sheet.name,sheet_coord[0],sheet_coord[1]))
            val = default

        return val


class UnitCurveCommand(FeatureCurveCommand):
    """
    Measures tuning curve(s) of particular unit(s).
    """

    pattern_presenter = param.Callable(
        default=PatternPresenter(pattern_generator=SineGrating(mask_shape=Disk(smoothing=0.0,size=1.0)),
                                 contrast_parameter="weber_contrast"))

    size=param.Number(default=0.5,bounds=(0,None),doc="""
        The size of the pattern to present.""")
    
    coords = param.List(default=[(0,0)],doc="""
        List of coordinates of units to measure.""")

    __abstract = True


__all__ = [
    "DistributionMatrix",
    "FullMatrix",
    "FeatureResponses",
    "ReverseCorrelation",
    "FeatureMaps",
    "FeatureCurves",
    "Feature",
    "PatternPresenter",
    "Subplotting",
    "MeasureResponseCommand",
    "SinusoidalMeasureResponseCommand",
    "PositionMeasurementCommand",
    "SingleInputResponseCommand",
    "FeatureCurveCommand",
    "UnitCurveCommand",
]



"""
EventProcessor classes.

An EventProcessor is an object that the Simulation is aware of, and can
accept simulation Events. EventProcessors can also generate Events, and
will presumably do some computation as well.  Most EventProcessors
will be in other more specific packages, such as topo.sheet; those
here are the remaining uncategorized EventProcessors.  Any new
EventProcessors classes added to this directory will automatically
become available for any model.

$Id$
"""

import param

from topo.base.simulation import EventProcessor,EPConnectionEvent


class PulseGenerator(EventProcessor):
    """
    A simple pulse generator node.

    Produces pulses (scalars) of a fixed amplitude at a fixed
    frequency and phase.  Period and phase are in units of simulation
    time.
    """
    
    dest_ports=None # Allows connections to come in on any port

    amplitude = param.Number(1.0,doc="The size of the pulse to generate.")
    
    period    = param.Number(1.0,bounds=(0.0,None),doc=
        "The period with which to repeat the pulse. Must be greater than zero.")
    
    phase     = param.Number(0.0,doc=
        "The time after starting the simulation to wait before sending the first pulse.")

    def input_event(self,conn,data):
        """On input from self, generate output. Ignore all other inputs."""
        self.verbose("Received event from ",conn.src,'on port',conn.dest_port,'with data',data)
        self.send_output(data=self.amplitude)

    def start(self):
        conn=self.simulation.connect(self.name,self.name,delay=self.period)
        e=EPConnectionEvent(self.simulation.time()+self.simulation.convert_to_time_type(self.phase), conn)
        self.simulation.enqueue_event(e)
        EventProcessor.start(self)


class ThresholdUnit(EventProcessor):
    """
    A simple pulse-accumulator threshold node.  Accumulates incoming
    pulses.  When the accumulated value rises above threshold, it
    generates a pulse of a given amplitude and resets the accumulator
    to zero.
    """
    
    dest_ports=None # Allows connections to come in on any port

    threshold     = param.Number(default=1.0,doc="The threshold at which to fire.")
    initial_accum = param.Number(default=0.0,doc="The initial accumulator value.")
    amplitude     = param.Number(default=1.0,doc="The size of the pulse to generate.")

    def __init__(self,**params):
        EventProcessor.__init__(self,**params)
        self.accum = self.initial_accum

    def input_event(self,conn,data):
        if conn.dest_port == 'input':
            self.accum += data
            self.verbose("Received ",data,"; accumulator now",self.accum)
            if self.accum > self.threshold:
                self.send_output(data=self.amplitude)
                self.accum = 0
                self.verbose(`self` + ' firing, amplitude = ' + `self.amplitude`)


class SumUnit(EventProcessor):
    """A simple unit that outputs the running sum of input received thus far."""
    
    dest_ports=None # Allows connections to come in on any port

    def __init__(self,**params):
        super(SumUnit,self).__init__(**params)
        self.value = 0.0

    def input_event(self,conn,data):
        self.value += data
        self.debug("received",data,"from",conn.src,"value =",self.value)

    def process_current_time(self):
        self.debug("process_current_time called, time =",self.simulation.time(),"value =",self.value)
        if self.value:
            self.debug("Sending output:",self.value)
            self.send_output(data=self.value)
            self.value = 0.0

import re,os
__all__ = list(set(k for k,v in locals().items()
                    if isinstance(v, type)
                    and issubclass(v, EventProcessor)
                    and v is not EventProcessor
                    and not v.__name__.startswith('_')))
# Automatically discover all .py files in this directory. 
__all__ += [re.sub('\.py$','',f) for f in os.listdir(__path__[0])
           if re.match('^[^_].*\.py$',f)]




"""
Topographica SheetView objects and its subclasses.

For use with the Topographica plotting and data analysis mechanisms.
A Sheet object has its internal data which remains hidden, but it will
create views of this data depending on the Sheet defaults or the
information requested.  This way there can be multiple views recorded
for a single sheet, and a view can be passed around independent of the
originating source object.

$Id$
"""

import types
import operator

import param

import sheet

### JABHACKALERT!
###
### Based on the comments, this file is not yet completed.  E.g. are
### any of these operators implemented?  They should either be tested
### or removed.  Similarly for other features below -- they need to be
### either completed or removed.

# Valid operations that a SheetView should support for the list of
# sheets passed in.  There are other things that would be useful
# such as masking which will require a bit more work.  !!UNTESTED!!
ADD      = 'ADD'
SUBTRACT = 'SUB'
MULTIPLY = 'MUL'
DIVIDE   = 'DIV'
# These 'operator' functions require 2 parameters when using 'apply'.
operations = {ADD : operator.add,
              SUBTRACT : operator.sub,
              MULTIPLY : operator.mul,
              DIVIDE : operator.truediv}


class SheetView(param.Parameterized):
    """
    A SheetView is constructed from a matrix of values, a bounding box
    for that matrix, and a name.  There are two major ways to create a
    SheetView: one is from a single matrix of data from a single
    sheet, the other is by combining the matrices from multiple
    matrices or SheetViews.
    """

    cyclic = param.Boolean(default=False,doc=
        """Whether or not the values in this View's matrix represent a cyclic dimension.""")

    cyclic_range = param.Parameter(None,doc="""If cyclic is True, this value is the cyclic range.""")
    
    ### JCALERT! term_1 and term_2 should be more explicit...
    ### the 3 cases described in the doc, are they really useful?
    ### shouldn't it be simplified?
    def __init__(self, (term_1, term_2), src_name=None, precedence = 0.0, timestamp = -1, row_precedence = 0.5, **params):
        """
        For ``__init__(self, input_tuple, **params)``, there are three
        types of input_tuples::
        
        (matrix_data, matrix_bbox)
        
            This form locks the value of the sheetview to a single matrix.
            Terminating case of a composite SheetView.
            
        (operation, [tuple_list])
        
            'operation' is performed on the matrices collected from
            tuple_list.  See the list of valid operations in
            operations.keys().
             
            Each tuple in the tuple_list is one of the following::
            
                (SheetView, None)
                    Another SheetView may be passed in to create nested plots.
                (matrix_data, bounding_box)
                    Static matrix data complete with bounding box.
                (Sheet, sheet_view_name)
                    This gets sheet_name.sheet_view(sheet_view_name) each time
                    the current SheetView has its data requested by .view().

        (Sheet, sheet_view_name)
        
                Degenerate case that will pull data from another SheetView
                and not do any additional processing.  Don't yet know a
                use for this case, but documented for possible future use.
        """
        super(SheetView,self).__init__(**params)

        self.src_name = src_name
        self.precedence = precedence
        self.row_precedence = row_precedence
        self.timestamp = timestamp
        
        # Assume there's no such thing as an operator that can be mistaken
        # for a matrix_data element.  This is true as long as the real
        # values of the operator keys are strings.
        if term_1 not in operations.keys():      
            self.operation = ADD          
            self._view_list = [(term_1, term_2)]
        else:
            self.operation = operations[term_1]
            self._view_list = term_2
            if not isinstance(self._view_list,types.ListType):
                self._view_list = [self._view_list]
            

    def view(self):
        """
        Return the requested view as a (matrix, bbox) tuple.

        If the constructor was given multiple maps, the view must be
        built before being returned, which may lock in new views of
        data from the specified sheets.

        Inputs are in the variable self._view_list which is a list of
        tuples, with each tuple being a matrix and a bounding box, or
        a sheet and a map name.  The sequence cannot be dumped into
        maps just once because the raw maps may have changed, so other
        sheets must be queried for the data repeatedly.
        """
        maps = []
        for tup in self._view_list:
            (term_1, term_2) = tup
            if isinstance(term_1,sheet.Sheet):
                maps.append(term_1.sheet_views[term_2])
            elif isinstance(term_1,SheetView):
                maps.append(term_1.view())         # Don't care about term_2
            else:
                maps.append((term_1, term_2))      # Assume it is a matrix

        # Convert the list of (matrix, bbox) tuples into a single
        # matrix and another bounding box.
        return self.sum_maps(maps)


    def sum_maps(self,maps):
        """
        Convert the list of (matrix, bbox) tuples into a single matrix
        and another bounding box.  Not a protected function as it could
        prove useful to other areas of the simulator.

        THIS MUST BE EXPANDED IN THE FUTURE TO MAKE PROPER USE OF THE
        BOUNDING BOX INFORMATION. CURRENT (8/04) IMPLEMENTATION
        PASSES THE FIRST MAP IN THE LIST AS THE BOUNDING BOX FOR THE
        CONSTRUCTED VIEW.

        WOULD HAVE DONE AN ADD/INTERSECTION/UNION, BUT BOUNDINGREGION
        DOES NOT YET SUPPORT SUCH OPERATIONS.
        """
        result = maps.pop(0)
        for m in maps:
            # Needs to be changed
            result = (apply(self.operation, (result[0], m[0])), result[1]) 
        return result



class UnitView(SheetView):
    """
    SPRING 2005: Currently does not extend any functions, but can do
    so to add outlines, and other interesting features.

    Consists of an X,Y position for the unit that this View is
    created for.  Subclasses the SheetView class.

    UnitViews should be stored in Sheets via a tuple
    ('Weights',Sheet,Projection,X,Y).  The dictionary in Sheets can be
    accessed by any valid key.
    """

    ### JCALERT! UnitView (as well as SheetView have to be reviewed.
    def __init__(self, term_tuple, x, y, projection, timestamp, **params):
        """
        Subclass of SheetView.  Contains additional x,y member data.
        """
        super(UnitView,self).__init__(term_tuple, projection.src.name, projection.src.precedence, timestamp = timestamp, row_precedence = projection.src.row_precedence, **params)
        self.x = x
        self.y = y
        self.projection = projection


class ProjectionView(SheetView):
    """
    ProjectionViews should be stored in Sheets via a tuple
    ('Weights',Sheet,Projection). 
    """

    def __init__(self, term_tuple,projection,timestamp,**params):
        """
        Subclass of SheetView. 
        """
        super(ProjectionView,self).__init__(term_tuple, projection.src.name, projection.src.precedence, timestamp, row_precedence = projection.src.row_precedence, **params)
        self.projection = projection






"""
High-level interface to the Player client libraries.

The Player client libraries allow Python code to communicate with
hardware devices such as robots, cameras, and range sensors.

This is a temporary home for this file until it finds a permanent home
(maybe in the PlayerStage project or in PLASTK?)

$Id$
"""


import time,array
import playerc

from threading import RLock, Thread
from Queue import Queue

from operator import eq,ne
from math import pi


# JPALERT: Because of the global interpreter lock in Python, using
# Python threads (via the 'thread' or 'threading' modules does not
# necessarily provide low-latency polling of the player process.  In
# particular, long-running native functions (e.g. C/C++ foreign
# functions) will not be pre-empted, so the polling loop won't
# get a timeslice until they complete.  A better solution, especially
# on multicore machines, is true preemptive multiprocessing.  The
# 'processing' module should provide that, but it doesn't yet work
# properly on MacOS, and I haven't tested it yet on linux.  God knows
# what will happen on Windows.

def use_processing():
    """
    Configure the module to use the processing library for asynchronous
    process support. Use of the processing library requires the use of
    queues for communication with robot devices.
    """
    import processing
    global RLock, Thread, Queue
    RLock = processing.RLock  # pyflakes:ignore (optional alternative)
    Thread = processing.Process  # pyflakes:ignore (optional alternative)
    Queue = processing.Queue  # pyflakes:ignore (optional alternative)

def use_threading():
    """
    Configure the module to use the threading library for asynchronous
    process support. (the default)
    """
    import threading
    global RLock, Thread, Queue
    RLock = threading.RLock  # pyflakes:ignore (optional alternative)
    Thread = threading.Thread  # pyflakes:ignore (optional alternative)
    Queue = Queue.Queue  # pyflakes:ignore (optional alternative)

# JPALERT This is a HACK for the CVS version of Player, this value
# should be defined in the playerc module:
playerc.PLAYERC_OPEN_MODE = 1


class PlayerException(Exception):
    pass
    

def player_fn(error_op=ne,error_val=0):
    """
    Player function decorator.  Adds error checking.

    Takes an operator and a value, and compares the result
    of the function call with the value using the operator.
    If the result is true, a PlayerException is raised.  The
    default error condition is error_op = ne, error_value = 0,
    which raises an exception if fn(*args) != 0.
    """
    def wrap(fn):
        def new_fn(*args):
            if error_op(fn(*args),error_val):
                raise PlayerException(playerc.playerc_error_str())
        return new_fn
    return wrap


def synchronized(lock):
    """
    Simple synchronization decorator.

    Takes an existing lock and synchronizes a function or
    method on that lock.  Code taken from the Python Wiki
    PythonDecoratorLibrary:

    http://wiki.python.org/moin/PythonDecoratorLibrary
    """
    def wrap(f):
        def newFunction(*args, **kw):
            lock.acquire()
            try:
                return f(*args, **kw)
            finally:
                lock.release()
        return newFunction
    return wrap


def synched_method(f):
    """
    Synchronized method decorator.

    Like synchronized() decorator, except synched_method assumes
    that the first argument of the function is an instance containing
    a Lock object, and this lock is used for synchronization.
    """
    def newFunction(self,*args,**kw):
        self._lock.acquire()
        try:
            return f(self,*args, **kw)
        finally:
            self._lock.release()
    return newFunction
    


class PlayerObject(object):
    """
    A generic threadsafe wrapper for client and proxy objects
    from the playerc library.

    PlayerObject wrappers are constructed automatically by PlayerRobot
    objects.  Each PlayerObject instance wraps a playerc device proxy
    or client object, and publishes a thread-safe version of each of
    proxy's methods, that is synchronized with the PlayerRobot
    instance's run-loop thread, and that catches playerc error
    conditions and raises them as PlayerExceptions.  The original
    playerc proxy object is available via the attribute .proxy.
    Specialized subclasses of PlayerObject can have additional
    interfaces for getting device state or setting commands specific
    to that device.

    Developer note: the PlayerObject base class __init__ method
    automatically wraps each method on the proxy that (a) doesn't
    begin with '__' and (b) is not already in dir(self).  This way,
    subclasses can override the wrapping process by defining their own
    wrappers *before* the base class __init__ method is called.
    """
    def __init__(self,proxy,lock):

        self._lock = lock
        self.proxy = proxy
        for name in dir(proxy):
            attr = getattr(proxy,name)
            if name not in dir(self) and name[:2] != '__' and callable(attr):
                setattr(self,name,synchronized(lock)(player_fn()(attr)))

        self.cmd_queue = Queue()

    def process_queues(self):
        while not self.cmd_queue.empty():
            name,args = self.cmd_queue.get()
            try:
                print "Doing command:",name,args
                getattr(self,name)(*args)
            finally:
                self.cmd_queue.task_done()


class PlayerClient(PlayerObject):
    """
    Player object wrapper for playerc.client objects.
    """
    def __init__(self,proxy,lock):

        # Override the wrapper on playerc.client.read, because
        # it returns None for errors, instead of returning 0 for
        # "no error."
        self.read = synchronized(lock)(player_fn(eq,None)(proxy.read))
        super(PlayerClient,self).__init__(proxy,lock)

    def process_queues(self):
        pass


class PlayerDevice(PlayerObject):
    """
    Generic Player device object.

    Overrides the default proxy .subscribe method so that the mode defaults
    to PLAYERC_OPEN_MODE.
    """

    @synched_method
    @player_fn()
    def subscribe(self,mode=playerc.PLAYERC_OPEN_MODE):
        return self.proxy.subscribe(mode)



class PTZDevice(PlayerDevice):
    """
    Player Pan/Tilt/Zoom (PTZ) device.

    Adds the following to the original proxy interface:

    state = The tuple (pan,tilt,zoom) indicating the current state of
    the PTZ device.

    state_deg = Same as state, but returns values in  degrees instead
    of radians

    set_deg() and set_ws_deg() methods.  Same as .set() and .set_ws(),
    using degrees instead of radians.
    """
    def get_state(self):
        return self.proxy.pan, self.proxy.tilt, self.proxy.zoom
    state = property(get_state)

    def get_state_deg(self):
        return self.proxy.pan*180/pi, \
               self.proxy.tilt*180/pi, \
               self.proxy.zoom*180/pi
    state_deg = property(get_state_deg)
    
    def set_deg(self,pan,tilt,zoom):
        self.set(pan*pi/180, tilt*pi/180, zoom*pi/180)
        
    def set_ws_deg(self,pan,tilt,zoom,pan_speed,tilt_speed):
        self.set_ws(pan*pi/180, tilt*pi/180, zoom*pi/180,pan_speed*pi/180,tilt_speed*pi/180)



class CameraDevice(PlayerDevice):
    """
    A Player camera device.

    The synchronized method get_image grabs an uncompressed snapshot,
    along with the additional formatting information needed to make an
    image.
    """

    def __init__(self,proxy,lock):
        self.decompress = synchronized(lock)(player_fn(ne,None)(proxy.decompress))
        super(CameraDevice,self).__init__(proxy,lock)
        self.image_queue = Queue()

    def process_queues(self):
        im = self.image
        # check to make sure it's really an image
        if im[1] > 0:
            self.image_queue.put(im)
        super(CameraDevice,self).process_queues()

        
#    @synched_method
    def get_image(self):
        """
        Returns the tuple:
          (format,width,height,bpp,fdiv,data)
        Where data is a copy of the uncompressed image data.
        """
        if self.proxy.compression:
            self.decompress()
        im_array = array.array('B')
        im_array.fromstring(self.proxy.image[:self.proxy.image_count])
        return self.proxy.format, \
               self.proxy.width,  \
               self.proxy.height, \
               self.proxy.bpp,    \
               self.proxy.fdiv,   \
               im_array

    image = property(get_image)
    


##################                    
# DEVICE TABLE
#
# This table contains the mapping from device type names
# to specialized device object types.  Types not indexed in this table
# should default to type PlayerDevice.

device_table = {'ptz'     :PTZDevice,
                'camera'  :CameraDevice,
                }



class PlayerRobot(object):
    """
    Player Robot interface.

    A PlayerRobot instance encapsulates an interface to a Player
    robot. It creates and manages a playerc.client object and a set of
    device proxies wrapped in PlayerDevice objects.  In addition, it
    maintains a run-loop in a separate thread that calls the client's
    .read() method at regular intervals.  The devices are published
    through standard interfaces on the PlayerRobot instance, and their
    methods and properties are synchronized with the run thread
    through a mutex.

    Example:

      # set up a robot object with position, laser, and camera objects
      robot = PlayerRobot(host='myrobot.mydomain.edu',port=6665,
                          devices = [('position2d',0),
                                     ('laser',0),
                                     ('camera',1)])

      # start the run thread, devices will be subscribed
      # automatically.
      robot.start()

      # start the robot turning at 30 deg/sec
      robot.position2d[0].set_cmd_vel(0, 0, 30*pi/180)

      # wait for a while
      time.sleep(5.0)

      # all stop
      robot.position2d[0].set_cmd_vel(0,0,0)

      # shut down the robot's thread, unsubscribing all devices and
      # disconnecting the client
      robot.stop()
    """
    
    def __init__(self,host='localhost',port=6665,speed=20,
                 devices=[]):

        self._thread = None
        self._running = False
        self._lock = RLock()
        self.speed = speed

        self._client = PlayerClient(playerc.playerc_client(None,host,port),self._lock)

        self._queues_running = False
        self._devices = []
        for devname,devnum in devices:
            self.add_device(devname,devnum=devnum)

    def start(self):
        assert self._thread is None
        self._thread = Thread(target=self.run_loop,name="PlayerRobot Run Loop")
        self._thread.setDaemon(True)
        self._thread.start()
        
    def stop(self):
        self._running = False
        self._thread.join()
        self._thread = None
        
    def run_loop(self):
        self._client.connect()
        self._running = True
        self.subscribe_all()
        try:
            while self._running:
                self._client.read()
                if self._queues_running:
                    self.process_queues()
                time.sleep(1.0/self.speed)
        finally:
            self.unsubscribe_all()
            self._client.disconnect()

    def run_queues(self,run_state):
        """
        When using queues for communication with devices, this method
        toggles queue processing.  It is often useful to turn off
        queue processing, e.g. when a client does not plan on using queued
        data for a while.
        """
        self._queues_running = run_state
        
    def process_queues(self):
        for d in self._devices:
            d.process_queues()

    def subscribe_all(self):
        for dev in self._devices:
            dev.subscribe()

    def unsubscribe_all(self):
        for dev in self._devices:
            dev.unsubscribe()

    def add_device(self,devname,devnum=0):
        if devname not in dir(self):
            setattr(self,devname,{})

        proxy_constr =  getattr(playerc,'playerc_'+devname)
        devtype = device_table.get(devname,PlayerDevice)
        
        try:
            self._lock.acquire()
            dev = devtype(proxy_constr(self._client.proxy,devnum),self._lock)
        finally:
            self._lock.release()
            
        self._devices.append(dev)
        getattr(self,devname)[devnum] = dev
        if self._running:
            dev.subscribe()
            



"""
Hierarchy of PlotGroup classes, i.e. output-device-independent sets of
plots.

Includes PlotGroups for standard plots of anything in a Sheet
database, plus weight plots for one unit, and projections.

$Id$
"""

import copy
import Image

import param
from param.parameterized import ParamOverrides
from param import resolve_path

import topo

from topo.base.cf import CFSheet,CFProjection
from topo.base.projection import ProjectionSheet
from topo.sheet import GeneratorSheet,Sheet
from topo.misc.keyedlist import KeyedList

from plot import make_template_plot, Plot
from plotfilesaver import PlotGroupSaver,CFProjectionPlotGroupSaver

# General CEBALERTs for this file:
# * It is very difficult to understand what is happening in these
#   classes!
# * Improve hierarchy. E.g. GridPlotGroup should be a mixin?
# * Oh yeah, and don't forget about the TestPattern window. It has
#   some kind of PlotGroup in it too...
# * There are missing classes (e.g. currently only the CFProjections
#   can be plotted for CFSheets).
# * There are no unit tests


def _cmp_plot(plot1,plot2):
    """
    Comparison function for sorting Plots.
    It compares the precedence number first and then the src_name and name attributes.
    """
    if plot1.precedence != plot2.precedence:
        return cmp(plot1.precedence,plot2.precedence)
    else:
        return cmp((plot1.plot_src_name+plot1.name),
                   (plot2.plot_src_name+plot2.name))


class PlotGroup(param.Parameterized):
    """
    Container that has one or more Plots and also knows how to arrange
    the plots and other special parameters.
    """

    pre_plot_hooks = param.HookList(default=[],doc="""
        Commands to execute before updating this plot, e.g. to calculate sheet views.
        
        The commands can be any callable Python objects, i.e. any x for
        which x() is valid.  The initial value is determined by the
        template for this plot, but various arguments can be passed, a
        modified version substituted, etc.""")

    plot_hooks = param.HookList(default=[],doc="""
        Commands to execute when redrawing a plot rather than regenerating data.

        E.g, for a plot with data measured once but displayed one
        sheet or unit at at time, this command will be called whenever
        the sheet or coordinate of unit to be plotted (or the
        simulator time) has changed.
        
        The commands can be any callable Python objects, i.e. any x for
        which x() is valid.  The initial value is determined by the
        template for this plot, but various arguments can be passed, a
        modified version substituted, etc.""")

    # I guess the interface for users of the class (I just mean methods
    # likely to be used) is:
    # make_plots()   
    # scale_images()  
    # + parameters
    
    # And the interface for subclasses (I mean methods likely to be
    # overridden) is:
    # _generate_plots()      - return the list of plots
    # _generate_labels()     - return the list of labels
    # _sort_plots()          - sort the list of plots
    # _exec_pre_plot_hooks() - run the pre_plot_hooks
    # _exec_plot_hooks()     - run the plot_hooks


    ##############################
    ########## interface for users

    def make_plots(self,update=True):
        """
        Create and scale the plots, after first executing the PlotGroup's pre_plot_hooks
        (if update is True) and plot_hooks.
        """
        if update:self._exec_pre_plot_hooks()
        self._exec_plot_hooks()
        self._create_images(update)
        self.scale_images()

    def scale_images(self,zoom_factor=None):
        """Scale the images by the given zoom factor, if appropriate; default is to do nothing."""
        pass


    ###################################
    ########## interface for subclasses
    
    def _generate_plots(self):
        """Return the list of Plots"""
        # subclasses may have dynamically generated Plots to add
        return self._static_plots[:]


    def _generate_labels(self):
        return [plot.label() for plot in self.plots]


    def _sort_plots(self):
        """Sort plots according to their precedence, then alphabetically."""
        self.plots.sort(_cmp_plot)


    def __init__(self,**params):
        super(PlotGroup,self).__init__(**params)
        self._static_plots = []
        self.plots = []
        self.labels = [] 
        self.time = None
        self.filesaver = PlotGroupSaver(self)

        # In the future, it might be good to be able to specify the
        # plot rows and columns using tuples.  For instance, if three
        # columns are desired with the plots laid out from left to
        # right, we could use (None, 3).  If three rows are desired
        # then (3, None).  Another method that would work is [3,2,4]
        # would have the first row with 3, the second row with 2, the
        # third row with 4, etc.  The default left-to-right ordering
        # in one row could perhaps be represented as (None, Inf).
        # 
        # Alternatively, we could add another precedence value, so
        # that the usual precedence value controls where the plot
        # appears left to right, while a rowprecedence value would
        # control where it appears top to bottom.  All plots with the
        # same rowprecedence would go on the same row, and the actual
        # value of the rowprecedence would determine which row goes
        # first.
        

    def _exec_pre_plot_hooks(self,**kw):
        for f in self.pre_plot_hooks: 
            f(**kw)


    def _exec_plot_hooks(self,**kw):
        for f in self.plot_hooks: 
            f(**kw)


    # unlikely to be overridden?
    def _create_images(self,update):
        """
        Generate the sorted and scaled list of plots constituting the PlotGroup.
        """
        self.plots = [plot for plot in self._generate_plots() if plot is not None]

        # Suppress plots in the special case of plots not being updated 
        # and having no resizable images, to suppress plotgroups that
        # have nothing but a color key
        resizeable_plots = [plot for plot in self.plots if plot.resize]
        if not update and not resizeable_plots:
            self.plots=[]

        # Take the timestamps from the underlying Plots
        timestamps = [plot.timestamp for plot in self.plots
                      if plot.timestamp >= 0]
        if len(timestamps)>0:
            self.time = max(timestamps)
            if max(timestamps) != min(timestamps):
                self.warning("Combining Plots from different times (%s,%s)" %
                             (min(timestamps),max(timestamps)))

        self._sort_plots()      
        self.labels = self._generate_labels()



### In the rest of the file, whenever we do a loop through all the
### simulation's sheets, seems like we should have set the sheet
### Parameter's range instead. sheet's range is set by the GUI.
### i.e. maybe there should be a sheet parameter here, with its range
### set when the plotgroup is instantiated. Then if sheet=None, the
### sheet's range can be used as the list of sheets.

class SheetPlotGroup(PlotGroup):

    # CEBALERT: shouldn't this be abstract? Currently it wouldn't do
    # anything with its sheets if you asked it to make plots.
    #
    # (TemplatePlotGroup and TestPatternPlotGroup are the only two
    # immediate children at the moment.)

    sheet_coords = param.Boolean(default=False,doc="""
        Whether to scale plots based on their relative sizes in sheet
        coordinates.  If true, plots are scaled so that their sizes are
        proportional to their area in sheet coordinates, so that one can
        compare corresponding areas.  If false, plots are scaled to have
        similar sizes on screen, regardless of their corresponding
        sheet areas, which maximizes the size of each plot.""")

    # CEBALERT: in the GUI, this isn't getting the right help text. I
    # suspect I put in a hack for these parameters in parametersframe,
    # sp when used outside of a parametersframe, they don't get the
    # help text.
    normalize = param.ObjectSelector(default='None',
                                     objects=['None','Individually'],
                                  doc="""
        'Individually': scale each plot so that the peak value will be white
        and the minimum value black.  

        'None': no scaling - 0.0 will be black and 1.0 will be white.  

        Normalization has the advantage of ensuring that any data that
        is present will be visible, but the disadvantage that the
        absolute scale will be obscured.  Non-normalized plots are
        guaranteed to be on a known scale, but only values between 0.0
        and 1.0 will be visibly distinguishable.""")

    integer_scaling = param.Boolean(default=False,doc="""
        When scaling bitmaps, whether to ensure that the scaled bitmap is an even
        multiple of the original.  If true, every unit will be represented by a
        square of the same size.  Typically false so that the overall area will
        be correct, e.g. when using Sheet coordinates, which is often more
        important.""")

    auto_refresh = param.Boolean(default=False,doc="""
        If this plot is being displayed persistently (e.g. in a GUI),
        whether to regenerate it automatically whenever the simulation
        time advances.  The default is False, because many plots are
        slow to generate (including most preference map plots).""")

    desired_maximum_plot_height = param.Number(default=0,bounds=(0,None),doc="""
        User-specified height of the tallest plot in this PlotGroup.
        Other plots will generally be scaled as appropriate, either
        to match this size (when sheet_coords is False), or to
        have the appropriate relative size (when sheet_coords is True).

        If enforce_minimum_plot_height is True, the actual maximum
        plot height may be larger than this parameter's value.  In
        particular, with enforce_minimum_plot_height=True, the default
        value of 0 gives plots that are the size of the underlying
        matrix, which is the most efficient size for saving plots
        directly to disk.  Larger values (e.g. 150) are suitable when
        displaying plots on screen.""")
        
    enforce_minimum_plot_height = param.Boolean(default=True,doc="""
        If true, ensure that plots are never shown smaller than their
        native size, i.e. with fewer than one pixel per matrix unit.
        This option is normally left on for safety, so that no
        visualization will be missing any units.  However, it may be
        acceptable to turn this check off when working with matrix
        sizes much larger than your screen resolution.""")

    # For users, adds:
    # sheets()
    # update_maximum_plot_height()
    # scale_images() 

    #########################
    ########## adds for users

    def sheets(self):
        return topo.sim.objects(Sheet).values()


    def update_maximum_plot_height(self,zoom_factor=None):
        """
        Determine which plot will be the largest, based on the
        settings for minimum plot heights, the specified zoom factor
        (if not None), etc.
        
        A minimum size (and potentially a maximum size) are enforced,
        as described below.

        If the scaled sizes would be outside of the allowed range,
        False is returned.  

        For matrix coordinate plots (sheet_coords=False), the minimum
        size is calculated as the native size of the largest bitmap to
        be plotted.  Other plots are then usually scaled up to (but
        not greater than) this size, so that all plots are
        approximately the same size, and no plot is missing any pixel.

        For Sheet coordinate plots, the minimum plotting density that
        will avoid losing pixels is determined by the maximum density
        from any sheet.  If all plots are then drawn at that density
        (as they must be for them to be in Sheet coordinates), the
        largest plot will then be the one with the largest sheet
        bounds, and the size of that plot will be the maximum density
        times the largest sheet bounds.
        """
        # Apply optional scaling to the overall size
        if zoom_factor:
            self.desired_maximum_plot_height*=zoom_factor
            
        self.maximum_plot_height=self.desired_maximum_plot_height

        if (self.enforce_minimum_plot_height):
            resizeable_plots = [p for p in self.plots if p.resize]

            # CEBALERT: not sure how I should have handled this...
            if not resizeable_plots:
                return False

            if not self.sheet_coords:
                bitmap_heights = [p._orig_bitmap.height() for p in resizeable_plots]
                minimum_height_of_tallest_plot = max(bitmap_heights)
                
            else:           
                ### JABALERT: Should take the plot bounds instead of the sheet bounds
                ### Specifically, a weights plot with situate=False
                ### doesn't use the Sheet bounds, and so the
                ### minimum_height is significantly overstated.
                
                sheets = topo.sim.objects(Sheet)
                max_sheet_height = max([(sheets[p.plot_src_name].bounds.lbrt()[3]-
                                         sheets[p.plot_src_name].bounds.lbrt()[1])
                                       for p in resizeable_plots])
                max_sheet_density = max([sheets[p.plot_src_name].xdensity
                                         for p in resizeable_plots])
                minimum_height_of_tallest_plot = max_sheet_height*max_sheet_density
                self.max_sheet_height=max_sheet_height
                    
            if (self.maximum_plot_height < minimum_height_of_tallest_plot):
                self.maximum_plot_height = minimum_height_of_tallest_plot
                # Return failure if trying to reduce but hit the minimum
                if zoom_factor:
                    return zoom_factor>1.0

#       print zoom_factor, self.maximum_plot_height, self.desired_maximum_plot_height
        return True
            

    ### CEB: At least some of this scaling would be common to all
    ### plotgroups, if some (e.g. featurecurve) didn't open new
    ### windows.
    ###
    ### Could strip out the matrix-coord scaling and put it into
    ### PlotGroup
    def scale_images(self,zoom_factor=None):
        """
        Enlarge or reduce the bitmaps as needed for display.

        The calculated sizes will be multiplied by the given
        zoom_factor, if it is not None.
        
        If the scaled sizes would be outside of the allowed range, no
        scaling is done, and False is returned.  (One might
        conceivably instead want the scaling to reach the actual
        minimum or maximum allowed, but if we did this, then repeated
        enlargements and reductions would not be reversible, unless we
        were very tricky about how we did it.)
        """

        # No scaling to do if there are no scalable plots, or no desired size
        resizeable_plots = [p for p in self.plots if p.resize]
        if not resizeable_plots or not self.desired_maximum_plot_height:
            return False

        # Calculate maximum_plot_height, and abort if not feasible
        if not self.update_maximum_plot_height(zoom_factor) and zoom_factor:
            return False
            
        # Scale the images so that each has a size up to the self.maximum_plot_height
        for plot in resizeable_plots:
            if self.sheet_coords:
                s = topo.sim.objects(Sheet)[plot.plot_src_name]
                scaling_factor=self.maximum_plot_height/float(s.xdensity)/self.max_sheet_height
            else:
                scaling_factor=self.maximum_plot_height/float(plot._orig_bitmap.height())
            
            if self.integer_scaling:
                scaling_factor=max(1,int(scaling_factor))
                
            plot.set_scale(scaling_factor)
            
            #print "Scaled %s %s: %d->%d (x %f)" % (plot.plot_src_name, plot.name, plot._orig_bitmap.height(), plot.bitmap.height(), scaling_factor)

        return True



def _get_value_range(plots):
    """
    Helper function to return the (min,max) given by the value_ranges of
    the given plots.

    Return None if there were no plots with value_range.
    """
    mins = []
    maxs = []
    for plot in plots:
        if hasattr(plot,'value_range'):
            mins.append(plot.value_range[0])
            maxs.append(plot.value_range[1])

    if len(mins)==0:
        return None
    else:
        return (min(mins),max(maxs))


class TemplatePlotGroup(SheetPlotGroup):
    """
    Container that allows creation of different types of plots in a
    way that is independent of particular models or Sheets.

    A TemplatePlotGroup is constructed from a plot_templates list, an
    optional command to run to generate the data, and other optional
    parameters.
    
    The plot_templates list should contain tuples (plot_name,
    plot_template).  Each plot_template is a list of (name, value)
    pairs, where each name specifies a plotting channel (such as Hue
    or Confidence), and the value is the name of a SheetView (such as
    Activity or OrientationPreference).
        
    Various types of plots support different channels.  An SHC
    plot supports Strength, Hue, and Confidence channels (with
    Strength usually being visualized as luminance, Hue as a color
    value, and Confidence as the saturation of the color).  An RGB
    plot supports Red, Green, and Blue channels.  Other plot types
    will be added eventually.

    For instance, one could define an Orientation-colored Activity
    plot as::
        
      plotgroups['Activity'] =
          TemplatePlotGroup(name='Activity', category='Basic',
              pre_plot_hooks=[measure_activity],
              plot_templates=[('Activity',
                  {'Strength': 'Activity', 'Hue': 'OrientationPreference', 'Confidence': None})])
    
    This specifies that the final TemplatePlotGroup will contain up to
    one Plot named Activity per Sheet, although there could be no
    plots at all if no Sheet has a SheetView named Activity once
    'measure_activity()' has been run.  The Plot will be colored by
    OrientationPreference if such a SheetView exists for that Sheet,
    and the value (luminance) channel will be determined by the
    SheetView Activity.  This plot will be listed in the category
    'Basic' anywhere such categories are relevant (e.g. in the GUI).

    
    Here's a more complicated example specifying two different plots
    in the same PlotGroup::

      TemplatePlotGroup(name='Orientation Preference', category='Basic'
          pre_plot_hooks=[measure_or_pref.instance()],
          plot_templates=
              [('Orientation Preference',
                  {'Strength': None, 'Hue': 'OrientationPreference'}),
               ('Orientation Selectivity',
                  {'Strength': 'OrientationSelectivity'})])

    Here the TemplatePlotGroup will contain up to two Plots per Sheet,
    depending on which Sheets have OrientationPreference and
    OrientationSelectivity SheetViews.


    The function create_plotgroup provides a convenient way to define plots using
    TemplatePlotGroups; search for create_plotgroup elsewhere in the code to see
    examples. 
    """

    doc = param.String(default="",doc="""
        Documentation string describing this type of plot.""")

    plot_immediately=param.Boolean(False,doc="""
        Whether to call the pre-plot hooks at once or only when the user asks for a refresh.
 
        Should be set to true for quick plots, but false for those that take a long time
        to calculate, so that the user can change the hooks if necessary.""")
    
    prerequisites=param.List([],doc="""
        List of preference maps that must exist before this plot can be calculated.""")

    category = param.String(default="User",doc="""
        Category to which this plot belongs, which will be created if necessary.""")

    # CEBALERT: how to avoid repeating documentation?
    # CB: also, documentation for normalization types needs cleaning up.
    normalize = param.ObjectSelector(default='None',
                                     objects=['None','Individually','AllTogether'],doc="""
        
        'Individually': scale each plot so that its maximum value is
        white and its minimum value black.

        'None': no scaling (0.0 will be black and 1.0 will be white).

        'AllTogether': scale each plot so that the highest maximum value is
        white, and the lowest minimum value is black.

    
        Normalizing 'Individually' has the advantage of ensuring that
        any data that is present will be visible, but the disadvantage
        that the absolute scale will be obscured.  Non-normalized
        plots are guaranteed to be on a known scale, but only values
        between 0.0 and 1.0 will be visibly distinguishable.""")



    # Overrides:
    # _generate_plots() - supports normalize=='AllTogether'

    # For users, adds:
    # add_template()
    # add_static_image()

    # For subclasses, adds:
    # _template_plots()             - 
    # _make_template_plot()         - 
    # _kw_for_make_template_plots() -


    #####################
    ########## overridden

    def _generate_plots(self):
        if self.normalize=='AllTogether':
            range_ = _get_value_range(self._template_plots(range_=None))
        else:
            range_ = False
        return self._static_plots[:]+self._template_plots(range_=range_)


    def __init__(self,plot_templates=None,static_images=None,**params):
        super(TemplatePlotGroup,self).__init__(**params)
        self.plot_templates = KeyedList(plot_templates or [])
        # Add plots for the static images, if any
        for image_name,file_path in static_images or []:
            self.add_static_image(image_name,file_path)


    #########################
    ########## adds for users

    # JCALERT! We might eventually write these two functions
    # 'Python-like' by using keyword argument to specify each
    # channel and then get the dictionnary of all remaining
    # arguments.
    # 
    # JABALERT: We should also be able to store a documentation string
    # describing each plot (for hovering help text) within each
    # plot template.

    def add_template(self,name,specification_tuple_list):
        dict_={}
        for key,value in specification_tuple_list:
            dict_[key]=value
        self.plot_templates.append((name,dict_))

    add_plot = add_template # CEBALERT: should be removed when callers updated


    def add_static_image(self,name,file_path):
        """
        Construct a static image Plot (e.g. a color key for an Orientation Preference map).
        """
        image = Image.open(resolve_path(file_path))
        plot = Plot(image,name=name)
        self._static_plots.append(plot)
            


    ##############################
    ########## adds for subclasses

    def _template_plots(self,range_=False):
        # calls make_template_plot for all plot_templates for all kw returned
        # by _kw_for_make_template_plot!!
        template_plots = []
        for plot_template_name,plot_template in self.plot_templates:
            for kw in self._kw_for_make_template_plot(range_):
                template_plots.append(self._make_template_plot(plot_template_name,plot_template,**kw))
        return template_plots


    def _kw_for_make_template_plot(self,range_):
        # Return a list of dictionaries; for each dictionary,
        # _make_template_plot() will be called with that dictionary
        # as keyword arguments.
        #
        # Allows subclasses to control what range of things (sheets,
        # projections) _make_template_plot() is called over, and to
        # control what keyword arguments
        # (sheet=,proj_=,range_=,bounds=,x=,...) are supplied.
        return [dict(sheet=sheet,range_=range_) for sheet in self.sheets()]


    def _make_template_plot(self,plot_template_name,plot_template,**kw):
        return make_template_plot(plot_template,
                                  kw['sheet'].sheet_views,
                                  kw['sheet'].xdensity,
                                  kw['sheet'].bounds,
                                  self.normalize,
                                  # CEBALERT: after this class, p.t. name never used
                                  name=plot_template_name,
                                  range_=kw['range_'])


    
def default_measureable_sheet():
    """Returns the first sheet for which measure_maps is True (if any), or else the first sheet, for use as a default value."""

    sheets = [s for s in topo.sim.objects(Sheet).values()
              if hasattr(s,'measure_maps') and s.measure_maps]
    if len(sheets)<1:
        sheets = [s for s in topo.sim.objects(Sheet).values()]
    if len(sheets)<1:
        raise ValueError("Unable to find a suitable measureable sheet.")
    sht=sheets[0]
    if len(sheets)>1:
        param.Parameterized().message("Using sheet %s." % sht.name)
    return sht



class ProjectionSheetPlotGroup(TemplatePlotGroup):
    """Abstract PlotGroup for visualizations of the Projections of one ProjectionSheet."""
    __abstract = True

    # Class attribute; forms part of the key for sheet_views (see
    # the _key() method).
    keyname = "ProjectionSheet" 

    sheet = param.ObjectSelector(default=None,
                                 compute_default_fn=default_measureable_sheet,
                                 doc=
        """The Sheet from which to produce plots.""")

    normalize = param.ObjectSelector(default='None',
                                  objects=['None','Individually','AllTogether','JointProjections'],doc="""
        'Individually': scale each plot so that the peak value will be white
        and the minimum value black.

        'None': no scaling - 0.0 will be black and 1.0 will be white.  

        'AllTogether': scale each plot so that the peak value of all the plots
        is white, and the minimum value of all the plots will be
        black.

        'JointProjections': as 'Individually', except that plots produced from
        projections whose weights are jointly normalized will be
        jointly normalized.
        
        Normalization has the advantage of ensuring that any data that
        is present will be visible, but the disadvantage that the
        absolute scale will be obscured.  Non-normalized plots are
        guaranteed to be on a known scale, but only values between 0.0
        and 1.0 will be visibly distinguishable.""")


    sheet_type = ProjectionSheet

    # Overrides:
    # _exec*plot_hooks()           - passes sheet to hooks
    # sheet()                      - returns only a single sheet
    # _kw_for_make_template_plot() - adds projections (+ assumes one sheet) 
    # _template_plots()            - support joint normalization
    # _make_template_plot()        -
    # _generate_labels()           - 

    # Adds for users:
    # projections() - 

    # Adds for subclasses:
    # _key() - 
    # _kw_for_one_proj() - 
    # _check_sheet_type() - 
    # _check_data_exist() - 


    #########################
    ########## adds for users

    def projections(self):
        return self.sheet.projections().values()


    #####################
    ########## overridden

    def sheets(self):
        return [self.sheet]

    def _exec_pre_plot_hooks(self,**kw):
        self.params('sheet').compute_default()
        self._check_sheet_type()
        super(ProjectionSheetPlotGroup,self)._exec_pre_plot_hooks(sheet=self.sheet,**kw)


    def _exec_plot_hooks(self,**kw):
        super(ProjectionSheetPlotGroup,self)._exec_plot_hooks(sheet=self.sheet,**kw)


    def _kw_for_make_template_plot(self,range_):
        args = []
        for proj in self.projections():
            for d in self._kw_for_one_proj(proj):
                d['range_']=range_[proj.name]
                args.append(d)
        return args


    def _template_plots(self,range_=False):
        # all the extra processing is for normalize=='JointProjections'

        # {name:range_} for the projections
        named_ranges = dict.fromkeys([proj.name for proj in self.projections()],range_)

        if self.normalize=='JointProjections':
            ranges = {}
            for group_key in self.sheet._grouped_in_projections('JointNormalize').keys():
                if group_key is None:
                    ranges[group_key]=False
                else:
                    projlist = self.sheet._grouped_in_projections('JointNormalize')[group_key]
                    self._check_data_exist(projlist)

                    # hack: need a _kw_for_make_template() that works
                    # with a single range value and restricts
                    # projections to those with a specified key.
                    self._group_key = group_key
                    _orig = self._kw_for_make_template_plot
                    self._kw_for_make_template_plot = self._hack

                    plotlist = super(ProjectionSheetPlotGroup,self)._template_plots(range_=None)
                    self._kw_for_make_template_plot = _orig
                    ranges[group_key] = _get_value_range(plotlist) 

            for key,proj in self.__keyed_projections():
                named_ranges[proj.name] = ranges[key]

        return super(ProjectionSheetPlotGroup,self)._template_plots(range_=named_ranges)


    def _make_template_plot(self,plot_template_name,plot_template,**kw):#sheet,proj
        return make_template_plot(self._channels(plot_template,**kw),
                                  kw['proj'].src.sheet_views,
                                  kw['proj'].src.xdensity,
                                  None,
                                  self.normalize,
                                  name=kw['proj'].name,
                                  range_=kw['range_'])


    def _generate_labels(self):
        return ["%s%s"%(plot.name,
                        (("\n(from %s)" % plot.proj_src_name)
                         if hasattr(plot,"proj_src_name") else ""))
                 for plot in self.plots]


    ##############################
    ########## adds for subclasses

    def _key(self,**kw):
        # the key for sheet_views 
        return (self.keyname,self.sheet.name,kw['proj'].name)


    def _kw_for_one_proj(self,proj):
        # Return a list of dictionaries; for each dictionary,
        # _make_template_plot() will be called with that dictionary as
        # keyword arguments.
        return [dict(proj=proj)]


    # CB: this isn't really necessary for these classes
    # themselves. Right now we have it provide a useful error message
    # to users (which is useful).
    def _check_sheet_type(self):
        if self.sheet is not None and not isinstance(self.sheet,self.sheet_type):
            raise TypeError(
                "%s's sheet Parameter must be set to a %s instance (currently %s, type %s)." \
                %(self,self.sheet_type,self.sheet,type(self.sheet))) 


    def _check_data_exist(self,projlist):
        # for joint normalization, not all data required may be
        # measured and displayed by the PlotGroup (in subclasses that
        # don't display all projections at once).
        pass

    # unlikely to be overridden?
    def _channels(self,plotgroup_template,**kw):
        channels = copy.deepcopy(plotgroup_template)
        channels['Strength']=self._key(**kw)
        return channels


    # CEBALERT: used to temporarily override
    # _kw_for_make_template_plot(), to allow calculation of range for
    # joint normalization. A mess.
    def _hack(self,range_):
        for key,projlist in self.sheet._grouped_in_projections('JointNormalize'):
            if key==self._group_key:
                args = []
                for proj in projlist or self.projections():
                    for d in self._kw_for_one_proj(proj):
                        d['range_']=range_
                        args.append(d)
                return args
        raise


    def __keyed_projections(self):
        # helper method to return a list of (key,proj) pairs from self.sheet
        keys_and_projns = []
        for key,projlist in self.sheet._grouped_in_projections('JointNormalize'):
            for proj in projlist:
                keys_and_projns.append((key,proj))
        return keys_and_projns



class ProjectionSheetMeasurementCommand(param.ParameterizedFunction):
    """A callable Parameterized command for measuring or plotting a specified Sheet."""

    sheet = param.ObjectSelector(default=None,doc="""
        Name of the sheet to use in measurements.""")

    __abstract = True



class ProjectionActivityPlotGroup(ProjectionSheetPlotGroup):
    """Visualize the activity of all Projections into a ProjectionSheet."""

    keyname='ProjectionActivity' 

    #####################
    ########## overridden

    def _make_template_plot(self,plot_template_name,plot_template,**kw):
        return make_template_plot(self._channels(plot_template,**kw),
                                  kw['proj'].dest.sheet_views,
                                  kw['proj'].dest.xdensity,
                                  kw['proj'].dest.bounds,
                                  self.normalize,
                                  name=kw['proj'].name,
                                  range_=kw['range_'])


# CEBALERT: Total mess. should be a mix-in class, or something like
# that.  
class GridPlotGroup(ProjectionSheetPlotGroup):
    """
    A ProjectionSheetPlotGroup capable of generating coordinates on a 2D grid.
    """
    ### JPALERT: The bounds are meaningless for large sheets anyway.  If a sheet
    ### is specified in, say, visual angle coordinates (e.g., -60 to +60 degrees), then
    ### the soft min of 5.0/unit will still give a 600x600 array of CFs!
    ### Density should probably be specified WRT to sheet bounds,
    ### instead of per-unit-of-sheet.    
    density = param.Number(default=10.0,
                     softbounds=(5.0,50.0),doc="""
                     Number of units to plot per 1.0 distance in sheet coordinates""")

    normalize = param.ObjectSelector(default='None',
                                  objects=['None','Individually','AllTogether']) # joint is removed


    # Adds for subclasses:
    # generate_coords() - 

    # Overrides:
    # _sort_plots()
    # _kw_for_make_template_plot() - no projection
    # _key()                       - no projection
    # _make_template_plot()        - no projection

    #####################
    ########## overridden

    def _sort_plots(self):
        """Skips plot sorting to keep the generated order."""
        pass


    def _kw_for_make_template_plot(self,range_):
        args = []
        for x,y in self.generate_coords():
            x_center,y_center = self.sheet.closest_cell_center(x,y)
            args.append(dict(x=x_center,y=y_center,sheet=self.sheet,range_=range_))
        return args


    def _key(self,**kw):
        return (self.keyname,self.sheet.name,kw['x'],kw['y'])


    def _make_template_plot(self,plot_template_name,plot_template,**kw):
        return make_template_plot(self._channels(plot_template,**kw),
                                  self.input_sheet.sheet_views,
                                  self.input_sheet.xdensity,
                                  self.input_sheet.bounds,
                                  self.normalize,
                                  range_=kw['range_'])

    # (should *skip* the joint normalization additions in
    # projectionsheet for some subclasses (e.g. RFProjectionPlot), but
    # should not for other subclasses (e.g. ProjectionPlotGroup))
#    def _template_plots(self,range_=False):
#        return TemplatePlotGroup._dynamic_plot(self,range_)



    ##############################
    ########## adds for subclasses

    def generate_coords(self):
        """
        Evenly space out the units within the sheet bounding box, so
        that it doesn't matter which corner the measurements start
        from.  A 4 unit grid needs 5 segments.  List is in left-to-right,
        from top-to-bottom.
        """
        def rev(x): y = x; y.reverse(); return y

        (l,b,r,t) = self.sheet.bounds.lbrt()
        x = float(r - l) 
        y = float(t - b)
        x_step = x / (int(x * self.density) + 1)
        y_step = y / (int(y * self.density) + 1)
        l = l + x_step
        b = b + y_step
        coords = []
        self.proj_plotting_shape = (int(y * self.density),int(x * self.density))

        for j in rev(range(self.proj_plotting_shape[0])):
            for i in range(self.proj_plotting_shape[1]):
                coords.append((x_step*i + l, y_step*j + b))

        return coords



def default_input_sheet():
    """Returns the first GeneratorSheet defined, for use as a default value."""
    sheets=topo.sim.objects(GeneratorSheet).values()
    if len(sheets)<1:
        raise ValueError("Unable to find a suitable input sheet.")
    sht=sheets[0]
    if len(sheets)>1:
        param.Parameterized().message("Using input sheet %s." % sht.name)
    return sht



class RFProjectionPlotGroup(GridPlotGroup):
    keyname='RFs'
    input_sheet = param.ObjectSelector(default=None,compute_default_fn=default_input_sheet,
                                       doc="The sheet on which to measure the RFs.")


    # Overrides:
    # _template_plots()      - no projection
    # _exec_pre_plot_hooks() - 


    #####################
    ########## overridden

    def _template_plots(self,range_=False):
        return TemplatePlotGroup._template_plots(self,range_=range_)


    def _exec_pre_plot_hooks(self,**kw): # RFHACK
        self.params('input_sheet').compute_default()
        super(RFProjectionPlotGroup,self)._exec_pre_plot_hooks(input_sheet=self.input_sheet,**kw)


class TwoOrientationsPlotGroup( TemplatePlotGroup ):
    """Display with small segments the two most preferred orientations for each
    units in the sheet. Only orientation with significative selectivity are
    shown. Darker segments correspond to orientations with more selective
    response."""

    unit_size       = param.Number(default=25,bounds=(9,None),doc="box size of a single unit")

    # Overrides:
    # _make_template_plot	- use density argument slot to parse unit_size

    def _make_template_plot(self,plot_template_name,plot_template,**kw):
        return make_template_plot(plot_template,
                                  kw['sheet'].sheet_views,
                                  self.unit_size,
                                  kw['sheet'].bounds,
                                  self.normalize,
                                  name=plot_template_name,
                                  range_=kw['range_'])




# CEBALERT: haven't modified; might need updating. Doesn't seem to be
# a way to call it from the GUI - is it known to be broken?
class RetinotopyPlotGroup(TemplatePlotGroup):
    input_sheet = param.ObjectSelector(default=None,compute_default_fn=default_input_sheet,
                                       doc="The sheet on which to measure the RFs.")

    def _exec_pre_plot_hooks(self,**kw): # RFHACK
        self.params('input_sheet').compute_default()
        super(RetinotopyPlotGroup,self)._exec_pre_plot_hooks(
            input_sheet=self.input_sheet,**kw)



class ProjectionPlotGroup(GridPlotGroup):

    keyname='Weights'

    projection = param.ObjectSelector(default=None,doc="The projection to visualize.")

    # ProjectionSheetPlotGroup
    normalize = param.ObjectSelector(default='None',
                                  objects=['None','Individually','AllTogether','JointProjections'])


    # Overrides:
    # _projections()               - one projection
    # _key()                       - restores projection
    # _kw_for_make_template_plot() - restores projection
    # _exec_pre_plot_hooks()       - pass coords to hooks
    # _check_data_exist() - check for data from other jointly normalized projections 


    #####################
    ########## overridden

    def projections(self):
        return [self.projection]


    # GridPlotGroup+ProjectionSheetPlotGroup
    def _key(self,**kw):
        return (self.keyname,self.sheet.name,kw['proj'].name,kw['x'],kw['y'])


    # ProjectionSheetPlotGroup
    def _check_data_exist(self,projlist):
        nodata = []
        for proj in projlist:
            d = self._kw_for_one_proj(proj)[0] # only checking one (x,y)
            if self._key(**d) not in proj.src.sheet_views:
                nodata.append(proj.name)
        if len(nodata)>0:
            raise ValueError("Joint normalization cannot proceed unless data has been measured for all jointly normalized projections (no data for %s)"%nodata)


    # GridPlotGroup+ProjectionSheetPlotGroup
    def _kw_for_make_template_plot(self,range_):
        args = self._kw_for_one_proj(self.projection)
        for d in args:
            d['range_']=range_[d['proj'].name]
        return args


    # GridPlotGroup+ProjectionSheetPlotGroup
    def _exec_pre_plot_hooks(self,**kw):
        coords=self.generate_coords() # why not in gridplotgroup?
        super(ProjectionPlotGroup,self)._exec_pre_plot_hooks(coords=coords,projection=self.projection,**kw)



class UnitMeasurementCommand(ProjectionSheetMeasurementCommand):
    """A callable Parameterized command for measuring or plotting specified units from a Sheet."""

    coords = param.List(default=[(0,0)],doc="""
        List of coordinates of unit(s) to measure.""")

    projection = param.ObjectSelector(default=None,doc="""
        Name of the projection to measure; None means all projections.""")

    __abstract = True

    def __call__(self,**params):
        p=ParamOverrides(self,params)
        s = p.sheet
        if s is not None:
            for x,y in p.coords:
                s.update_unit_view(x,y,'' if p.projection is None else p.projection.name)



class CFProjectionPlotGroup(ProjectionPlotGroup):
    """Visualize one CFProjection."""

    situate = param.Boolean(default=False,doc="""
    If True, plots the weights on the entire source sheet, using zeros
    for all weights outside the ConnectionField.  If False, plots only
    the actual weights that are stored.""")
                     
    sheet_type = CFSheet

    # Overrides:
    # _make_template_plot()  - add bounds
    # _kw_for_one_proj()     - add bounds
    # _exec_pre_plot_hooks() - check proj type 

    # Adds for subclasses:
    # _check_projection_type() - 


    #####################
    ########## overridden

    def _make_template_plot(self,plot_template_name,plot_template,**kw):
        return make_template_plot(self._channels(plot_template,**kw),
                                  kw['proj'].src.sheet_views,
                                  kw['proj'].src.xdensity,
                                  kw['bounds'],
                                  self.normalize,
                                  name=kw['proj'].name,
                                  range_=kw['range_'])


    def _kw_for_one_proj(self,proj):
        args = []
        for x,y in self.generate_coords():            
            if self.situate:
                bounds = proj.src.bounds
            else:
                (r,c) = proj.dest.sheet2matrixidx(x,y)
                bounds = proj.cf_bounds(r,c)
            args.append(dict(x=x,y=y,proj=proj,bounds=bounds))

        return args


    def _exec_pre_plot_hooks(self,**kw): 
        self._check_projection_type()
        super(CFProjectionPlotGroup,self)._exec_pre_plot_hooks(**kw)


    def __init__(self,**params):
        super(CFProjectionPlotGroup,self).__init__(**params)    
        self.filesaver = CFProjectionPlotGroupSaver(self)


    ##############################
    ########## adds for subclasses

    # CB: same comment for ProjectionSheetPlotGroup's _check_sheet_type.
    def _check_projection_type(self):
        if not isinstance(self.projection,CFProjection):
            raise TypeError(
                "%s's projection Parameter must be set to a %s instance (currently %s, type %s)." \
                %(self,CFProjection,self.projection,type(self.projection))) 


# a mix-in? 
class UnitPlotGroup(ProjectionSheetPlotGroup):
    """
    Visualize anything related to a unit.
    """

    # JABALERT: need to show actual coordinates of unit returned
    x = param.Number(default=0.0,doc="""x-coordinate of the unit to plot""")
    y = param.Number(default=0.0,doc="""y-coordinate of the unit to plot""")
## """Sheet coordinate location desired.  The unit nearest this location will be returned.
## It is an error to request a unit outside the area of the Sheet.""")

    # Overrides:
    # _key()        - removes projection, adds coords
    # _exec*hooks() - includes coords 

    #####################
    ########## overridden

    def _key(self,**kw):
        return (self.keyname,self.sheet.name,self.x,self.y)

    def _exec_pre_plot_hooks(self,**kw):
        super(UnitPlotGroup,self)._exec_pre_plot_hooks(coords=[(self.x,self.y)],**kw)

    def _exec_plot_hooks(self,**kw):
        super(UnitPlotGroup,self)._exec_plot_hooks(coords=[(self.x,self.y)],**kw)



class ConnectionFieldsPlotGroup(UnitPlotGroup):
    """
    Visualize ConnectionField for each of a CFSheet's CFProjections.
    """
    keyname='Weights'

    sheet_type = CFSheet

    situate = param.Boolean(default=False,doc="""
    If True, plots the weights on the entire source sheet, using zeros
    for all weights outside the ConnectionField.  If False, plots only
    the actual weights that are stored.""")

    # Overrides:
    # _key()                - adds projection back
    # _make_template_plot() - adds projection back
    # _kw_for_one_proj()    - includes bounds


    #####################
    ########## overridden

    def _key(self,**kw):
        return (self.keyname,self.sheet.name,kw['proj'].name,self.x,self.y)


    def _make_template_plot(self,plot_template_name,plot_template,**kw):
        return make_template_plot(self._channels(plot_template,**kw),
                                  kw['proj'].src.sheet_views,
                                  kw['proj'].src.xdensity,
                                  kw['bounds'],
                                  self.normalize,
                                  name=kw['proj'].name,
                                  range_=kw['range_'])


    def _kw_for_one_proj(self,proj):
        if self.situate:
            bounds = None
        else:
            (r,c) = proj.dest.sheet2matrixidx(self.x,self.y)
            bounds = proj.cf_bounds(r,c)
        return [dict(proj=proj,bounds=bounds)]






class FeatureCurvePlotGroup(UnitPlotGroup):

    def _exec_pre_plot_hooks(self,**kw):
        super(FeatureCurvePlotGroup,self)._exec_pre_plot_hooks(**kw)
        self.get_curve_time()

    def _exec_plot_hooks(self,**kw):
        super(FeatureCurvePlotGroup,self)._exec_plot_hooks(**kw)
        self.get_curve_time()

    def get_curve_time(self):
        """
        Get timestamps from the current SheetViews in the curve_dict and
        use the max timestamp as the plot label
        Displays a warning if not all curves have been measured at the same time.
        """
        for x_axis in self.sheet.curve_dict.itervalues():
            for curve_label in x_axis.itervalues():
                timestamps = [SheetView.timestamp for SheetView in curve_label.itervalues()]

        if timestamps != []:
            self.time = max(timestamps)
            if max(timestamps) != min(timestamps):
                self.warning("Displaying curves from different times (%s,%s)" %
                             (min(timestamps),max(timestamps)))



plotgroups = KeyedList()
"""
Global repository of PlotGroups, to which users can add their own as
needed.
"""

plotgroup_types = {'Connection Fields': ConnectionFieldsPlotGroup,
                   'Projection': CFProjectionPlotGroup,
                   'RF Projection': RFProjectionPlotGroup,
                   'Retinotopy': RetinotopyPlotGroup,
                   'Projection Activity': ProjectionActivityPlotGroup}


def create_plotgroup(template_plot_type='bitmap',**params):
    """
    Create a new PlotGroup and add it to the plotgroups list.

    Convenience function to make it simpler to use the name of the
    PlotGroup as the key in the plotgroups list.

    template_plot_type: Whether the plots are bitmap images or curves
    ('curve').
    """
    name = params.get('name')

    pg_type = TemplatePlotGroup
    if name:
        if name in plotgroup_types:
            pg_type = plotgroup_types[name]
        elif template_plot_type=='curve':
            pg_type = FeatureCurvePlotGroup
                
    pg = pg_type(**params)
    plotgroups[pg.name]=pg
    return pg



"""
A family of response functions for CFProjections.

These function objects compute a response matrix when given an input
pattern and a set of ConnectionField objects.  Response functions come
in two varieties: ResponseFunction, and CFPResponseFunction.  A
ResponseFunction (e.g. DotProduct) computes the response due to one
CF.  To compute the response due to an entire CFProjection, a
ResponseFunction can be plugged in to CFPRF_Plugin.  CFPRF_Plugin is
one example of a CFPResponseFunction, which is a function that works
with the entire Projection at once.  Some optimizations and algorithms
can only be applied at the full CFPResponseFn level, so there are
other CFPResponseFns beyond CFPRF_Plugin.

Any new response functions added to this directory will automatically
become available for any model.

$Id$
"""

from topo.base.functionfamily import ResponseFn

# Imported here so that all ResponseFns will be in the same package
from topo.base.cf import DotProduct  # pyflakes:ignore (API import)

_public = list(set([_k for _k,_v in locals().items()
                    if isinstance(_v,type) and issubclass(_v,ResponseFn)]))

# Automatically discover all .py files in this directory.
import os,fnmatch
__all__ = _public + [f.split('.py')[0] for f in os.listdir(__path__[0]) if fnmatch.fnmatch(f,'[!._]*.py')]
del f,os,fnmatch



"""
Plot class.

$Id$
"""


import copy
import numpy

from numpy.oldnumeric import zeros, ones, Float, divide
from math import sin, cos

import param

from topo.base.sheetcoords import SheetCoordinateSystem,Slice

from bitmap import HSVBitmap, RGBBitmap, Bitmap, DrawBitmap


### JCALERT!
### - Re-write the test file, taking the new changes into account.
### - I have to change the order: situate, plot_bb and (normalize)
### - There should be a way to associate the density explicitly
###   with the sheet_views, because it must match all SheetViews
###   in that dictionary.  Maybe as a tuple?
### - Fix the plot name handling along with the view_info sheetview attribute
### - Get rid of release_sheetviews.


class Plot(param.Parameterized):
     """
     Simple Plot object constructed from a specified PIL image.
     """

     staleness_warning=param.Number(default=10,bounds=(0,None),doc="""
       Time length allowed between bitmaps making up a single plot before warning.

       If the difference between the SheetView with the earliest
       timestamp and the one with the latest timestamp is larger
       than this parameter's value, produce a warning.
       """)
     
     def __init__(self,image=None,**params):
          super(Plot,self).__init__(**params)
          self._orig_bitmap = Bitmap(image)
          self.bitmap = self._orig_bitmap # Possibly scaled copy (at first identical)
          self.scale_factor=1.0
          self.plot_src_name = ''
          self.precedence = 0.0
          self.row_precedence = 0.5
          # If False, this plot should be left in its native size
          # pixel-for-pixel, (e.g. for a color key or similar static
          # image), rather than being resized as necessary.
          self.resize=False

          # Time at which the bitmaps were created
          self.timestamp = -1


     def rescale(self,scale_factor):
          """
          Change the size of this image by the specified numerical factor.

          The original image is kept as-is in _orig_bitmap; the scaled
          image is stored in bitmap.  The scale_factor argument is
          taken as relative to the current scaling of the bitmap.  For
          instance, calling scale(1.5) followed by scale(2.0) will
          yield a final scale of 3.0, not 2.0.
          """
          self.scale_factor *= scale_factor

          if (self._orig_bitmap):
              self.bitmap = copy.copy(self._orig_bitmap)
              self.bitmap.image = self._orig_bitmap.zoom(self.scale_factor)


     def set_scale(self,scale_factor):
          """
          Specify the numerical value of the scaling factor for this image.

          The original image is kept as-is in _orig_bitmap; the scaled
          image is stored in bitmap.  The scale_factor argument is
          taken as relative to the original size of the bitmap.  For
          instance, calling scale(1.5) followed by scale(2.0) will
          yield a final scale of 2.0, not 3.0.
          """
          self.scale_factor = scale_factor

          if (self._orig_bitmap):
              self.bitmap = copy.copy(self._orig_bitmap)
              self.bitmap.image = self._orig_bitmap.zoom(self.scale_factor)

     def label(self):
          """Return a label for this plot."""
          return self.plot_src_name + '\n' + self.name



def _sane_plot_data(channels,sheet_views):
     # CEBALERT: was sf.net tracker item 1860837
     # (Avoid plotting only hue+confidence for a weights plot.)
     s_chan = channels.get('Strength')
     if s_chan is not None and len(s_chan)>0 and s_chan[0]=='Weights':
          return channels['Strength'] in sheet_views
     else:
          return True

# JABALERT: How can we handle joint normalization, where a set of
# plots (e.g. a CFProjectionPlotGroup, or the jointly normalized
# subset of a ConnectionFields plot) is all scaled by the same amount,
# so that relative strengths can be determined?  Maybe we can have
# make_template_plot and the various TemplatePlot types accept a
# parameter 'range_only' that makes them simply calculate a pair
# (min,max) with the values to use for scaling, and then the caller
# (e.g. CFProjectionPlotGroup._create_plots) would run through
# everything twice, first to get the ranges, and then the next time it
# would supply an explicit range for scaling (overriding the default
# single-plot normalization)?  See the commented-out code for
# value_range below for a start. I *think* that would work, but maybe
# there is some simpler way?
def make_template_plot(channels,sheet_views,density=None,
                       plot_bounding_box=None,normalize='None',
                       name='None',range_=False):
     """
     Factory function for constructing a Plot object whose type is not yet known.

     Typically, a TemplatePlot will be constructed through this call, because
     it selects the appropriate type automatically, rather than calling
     one of the Plot subclasses automatically.  See TemplatePlot.__init__ for
     a description of the arguments.
     """
     if _sane_plot_data(channels,sheet_views):
          plot_types=[SHCPlot,RGBPlot,PalettePlot,MultiOrPlot]
          for pt in plot_types:
               plot = pt(channels,sheet_views,density,plot_bounding_box,normalize,
                         name=name,range_=range_)
               if plot.bitmap is not None or range_ is None:
                    # range_ is None means we're calculating the range
                    return plot
     
     param.Parameterized(name="make_template_plot").verbose('No',name,'plot constructed for this Sheet')
     return None



class TemplatePlot(Plot):
    """
    A bitmap-based plot as specified by a plot template (or plot channels).
    """

    # Not sure why, but this has to be a Parameter to avoid spurious complaints
    warn_time=param.Number(-2,precedence=-1,doc="Time last warned about stale plots")


    def __init__(self,channels,sheet_views,density,
                 plot_bounding_box,normalize,
                 range_=False,**params):
        """
        Build a plot out of a set of SheetViews as determined by a plot_template.
        
        channels is a plot_template, i.e. a dictionary with keys
        (i.e. 'Strength','Hue','Confidence' ...).  Each key typically
        has a string value naming specifies a SheetView in
        sheet_views, though specific channels may contain other
        types of information as required by specific Plot subclasses.
        channels that are not used by a particular Plot subclass will
        silently be ignored.

        sheet_views is a dictionary of SheetViews, generally (but
        not necessarily) belonging to a Sheet object.

        density is the density of the Sheet whose sheet_views was
        passed.

        plot_bounding_box is the outer bounding_box of the plot to
        apply if specified.  If not, the bounds of
        the smallest SheetView are used.

        normalize specifies how the Plot should be normalized: any
        value of normalize other than 'None' will result in normalization
        according to the value of the range argument: 

          range=(A,B) - scale plot so that A is 0 and B is 1

          range=False - scale plot so that min(plot) is 0 and 
                        max(plot) is 1 (i.e. fill the maximim 
                        dynamic range)
 
          range=None  - calculate value_range only

        
        name (which is inherited from Parameterized) specifies the name
        to use for this plot.
        """
        super(TemplatePlot,self).__init__(**params) 
        # for a template plot, resize is True by default
        self.resize=True
        self.bitmap = None
        

        self.channels = channels
        self.view_dict = copy.copy(sheet_views)
        # bounds of the situated plotting area 
        self.plot_bounding_box = plot_bounding_box


        ### JCALERT ! The problem of displaying the right plot name is still reviewed
        ### at the moment we have the plot_src_name and name attribute that are used for the label.
        ### generally the name is set to the plot_template name, except for connection
        # set the name of the sheet that provides the SheetViews
        # combined with the self.name parameter when creating the plot (which is generally
        # the name of the plot_template), it provides the necessary information for displaying plot label
        self._set_plot_src_name()

        
        # # Eventually: support other type of plots (e.g vector fields...) using
        # # something like:
        # def annotated_bitmap(self):  
        # enable other construction....


    def _get_matrix(self,key):
        """
        Retrieve the matrix view associated with a given key, if any.

        If the key is found in self.channels and the corresponding
        sheetview is found in self.view_dict, the view's matrix is
        returned; otherwise None is returned (with no error).
        If the sheet_view derives from a cyclic distribution, and it
        will be used as Hue, the matrix is normalized in range 0..1
        """
        sheet_view_key = self.channels.get(key,None)
        sv = self.view_dict.get(sheet_view_key, None)
        if sv == None:
            matrix = None
        else:
            view = sv.view()
            matrix = view[0]
            if key=='Hue' and sv.cyclic:
                matrix /= sv.cyclic_range

            # Calculate timestamp for this plot
            timestamp = sv.timestamp
            if timestamp >=0:
                if self.timestamp < 0:
                    self.timestamp = timestamp
                elif abs(timestamp - self.timestamp) > self.staleness_warning:
                    if TemplatePlot.warn_time != min(timestamp, self.timestamp):
                        self.warning("Combining SheetViews from different times (%s,%s) for plot %s; see staleness_warning" %
                                     (timestamp, self.timestamp,self.name))
                        TemplatePlot.warn_time = min(timestamp, self.timestamp)
        return matrix


    def _set_plot_src_name(self):
        """ Set the Plot plot_src_name. Called when Plot is created"""
        for key in self.channels:
            sheet_view_key = self.channels.get(key,None)
            sv = self.view_dict.get(sheet_view_key, None)
            if sv != None:
                 self.plot_src_name = sv.src_name
                 self.precedence = sv.precedence
                 self.row_precedence = sv.row_precedence
                 if hasattr(sv,'proj_src_name'):
                      self.proj_src_name=sv.proj_src_name


    ### JCALERT: This could be inserted in the code of get_matrix
    def _get_shape_and_box(self):
        """
        Sub-function used by plot: get the matrix shape and the bounding box
        of the SheetViews that constitue the TemplatePlot.
        """
        for name in self.channels.values():
                sv = self.view_dict.get(name,None)
                if sv != None:
                     shape = sv.view()[0].shape
                     box = sv.view()[1]    
    
        return shape,box


    # CEBALERT: needs simplification! (To begin work on joint
    # normalization, I didn't want to interfere with the existing
    # normalization calculations.)  Also need to update this
    # docstring.
    #
    # range=None  - calculate value_range; don't scale a
    # range=(A,B) - scale a so that A is 0 and B is 1
    # range=False - scale a so that min(array) is 0 and max(array) is 1 
    def _normalize(self,a,range_):
        """ 
        Normalize an array s to be in the range 0 to 1.0.
        For an array of identical elements, returns an array of ones
        if the elements are greater than zero, and zeros if the
        elements are less than or equal to zero.
        """
        if range_: # i.e. not False, not None (expecting a tuple)
             range_min = float(range_[0])
             range_max = float(range_[1])

             if range_min==range_max:
                  if range_min>0: 
                       resu = ones(a.shape)
                  else:
                       resu = zeros(a.shape)
             else:
                  a_offset = a - range_min
                  resu = a_offset/(range_max-range_min)
            
             return resu
        else:
             if range_ is None:
                  if not hasattr(self,'value_range'):
                       self.value_range=(a.min(),a.max())
                  else: 
                       # If normalizing multiple matrices, take the largest values
                       self.value_range=(min(self.value_range[0],a.min()),
                                         max(self.value_range[1],a.max()))
                  return None # (indicate that array was not scaled)
             else: # i.e. range_ is False
                  a_offset = a-a.min()
                  max_a_offset = a_offset.max()

                  if max_a_offset>0:
                       a = divide(a_offset,float(max_a_offset))
                  else:
                       if min(a.ravel())<=0:
                            a=zeros(a.shape,Float)
                       else:
                            a=ones(a.shape,Float)
                  return a


    ### JC: maybe density can become an attribute of the TemplatePlot?
    def _re_bound(self,plot_bounding_box,mat,box,density):

        # CEBHACKALERT: for Julien...
        # If plot_bounding_box is that of a Sheet, it will already have been
        # setup so that the density in the x direction and the density in the
        # y direction are equal.
        # If plot_bounding_box comes from elsewhere (i.e. you create it from
        # arbitrary bounds), it might need to be adjusted to ensure the density
        # in both directions is the same (see Sheet.__init__()). I don't know where
        # you want to do that; presumably the code should be common to Sheet and
        # where it's used in the plotting?
        #
        # It's possible we can move some of the functionality
        # into SheetCoordinateSystem.
        if plot_bounding_box.containsbb_exclusive(box):
             ct = SheetCoordinateSystem(plot_bounding_box,density,density)
             new_mat = zeros(ct.shape,Float)
             r1,r2,c1,c2 = Slice(box,ct)
             new_mat[r1:r2,c1:c2] = mat
        else:
             scs = SheetCoordinateSystem(box,density,density)
             s=Slice(plot_bounding_box,scs)
             s.crop_to_sheet(scs)
             new_mat = s.submatrix(mat)

        return new_mat
  
             
             


class SHCPlot(TemplatePlot):
    """
    Bitmap plot based on Strength, Hue, and Confidence matrices.

    Constructs an HSV (hue, saturation, and value) plot by choosing
    the appropriate matrix for each channel.
    """

    def __init__(self,channels,sheet_views,density,
                 plot_bounding_box,normalize,
                 range_=False,**params):
        super(SHCPlot,self).__init__(channels,sheet_views,density, 
                                   plot_bounding_box,normalize,**params)
        
        # catching the empty plot exception
        s_mat = self._get_matrix('Strength')
        h_mat = self._get_matrix('Hue')
        c_mat = self._get_matrix('Confidence') 

        # If it is an empty plot: self.bitmap=None
        if (s_mat==None and c_mat==None and h_mat==None):
            self.debug('Empty plot.')

        # Otherwise, we construct self.bitmap according to what is specified by the channels.
        else:

            shape,box = self._get_shape_and_box()                                 

            hue,sat,val = self.__make_hsv_matrices((s_mat,h_mat,c_mat),shape,normalize,range_)
            
            if range_ is None:
                 return ##############################


            if self.plot_bounding_box == None:
               self.plot_bounding_box = box
                            
            hue = self._re_bound(self.plot_bounding_box,hue,box,density)
            sat = self._re_bound(self.plot_bounding_box,sat,box,density)
            val = self._re_bound(self.plot_bounding_box,val,box,density)
            
            self.bitmap = HSVBitmap(hue,sat,val)

        self._orig_bitmap=self.bitmap
        

    def __make_hsv_matrices(self,hsc_matrices,shape,normalize,range_=False):
        """ 
        Sub-function of plot() that return the h,s,v matrices corresponding 
        to the current matrices in sliced_matrices_dict. The shape of the matrices
        in the dict is passed, as well as the normalize boolean parameter.
        The result specified a bitmap in hsv coordinate.
    
        Applies normalizing and cropping if required.
        """
        zero=zeros(shape,Float)
        one=ones(shape,Float)   

        s,h,c = hsc_matrices
        # Determine appropriate defaults for each matrix
        if s is None: s=one # Treat as full strength by default
        if c is None: c=one # Treat as full confidence by default
        if h is None:       # No color, gray-scale plot.
            h=zero
            c=zero

        # If normalizing, offset the matrix so that the minimum
        # value is 0.0 and then scale to make the maximum 1.0
        if normalize!='None':
             s=self._normalize(s,range_=range_)
             # CEBALERT: I meant False, right?
             c=self._normalize(c,range_=False)
            

        # This translation from SHC to HSV is valid only for black backgrounds;
        # it will need to be extended also to support white backgrounds.
        hue,sat,val=h,c,s
        return (hue,sat,val)




    
class RGBPlot(TemplatePlot):
  """
  Bitmap plot based on Red, Green, and Blue matrices.
  
  Construct an RGB (red, green, and blue) plot from the Red, Green,
  and Blue channels.
  """
  def __init__(self,channels,sheet_views,density,
               plot_bounding_box,normalize,
               range_=False,**params):

       super(RGBPlot,self).__init__(channels,sheet_views,density, 
                                   plot_bounding_box,normalize,**params)


       # catching the empty plot exception
       r_mat = self._get_matrix('Red')
       g_mat = self._get_matrix('Green')
       b_mat = self._get_matrix('Blue') 

       # If it is an empty plot: self.bitmap=None
       if (r_mat==None and g_mat==None and b_mat==None):
            self.debug('Empty plot.')
            # Otherwise, we construct self.bitmap according to what is specified by the channels.
       else:

            shape,box = self._get_shape_and_box()                                 

            red,green,blue = self.__make_rgb_matrices((r_mat,g_mat,b_mat),shape,
                                                      normalize,range_=range_)

            if range_ is None:
                 return ############################

            if self.plot_bounding_box == None:
               self.plot_bounding_box = box
                            
            red = self._re_bound(self.plot_bounding_box,red,box,density)
            green = self._re_bound(self.plot_bounding_box,green,box,density)
            blue = self._re_bound(self.plot_bounding_box,blue,box,density)
         
            self.bitmap = RGBBitmap(red,green,blue)
        
       self._orig_bitmap=self.bitmap

  def __make_rgb_matrices(self, rgb_matrices,shape,normalize,range_=False):
        """ 
        Sub-function of plot() that return the h,s,v matrices
        corresponding to the current matrices in
        sliced_matrices_dict. The shape of the matrices in the dict is
        passed, as well as the normalize boolean parameter.  The
        result specified a bitmap in hsv coordinate.
    
        Applies normalizing and cropping if required.
        """
        zero=zeros(shape,Float)

        r,g,b = rgb_matrices
        # Determine appropriate defaults for each matrix
        if r is None: r=zero 
        if g is None: g=zero 
        if b is None: b=zero 

        # CEBALERT: have I checked this works?
        if normalize!='None':
             r = self._normalize(r,range_=range_)
             g = self._normalize(g,range_=range_)
             b = self._normalize(b,range_=range_)

        return (r,g,b)
   




class PalettePlot(TemplatePlot):
     """
     Bitmap plot based on a Strength matrix, with optional colorization.

     Not yet implemented.

     When implemented, construct an RGB plot from a Strength channel,
     optionally colorized using a specified Palette.
     """
  
     def __init__(self,channels,sheet_views,density,
                  plot_bounding_box,normalize,**params):

          super(PalettePlot,self).__init__(channels,sheet_views,density, 
                                           plot_bounding_box,normalize,**params)

          ### JABHACKALERT: To implement the class: If Strength is present,
          ### ask for Palette if it's there, and make a PaletteBitmap.



             


class MultiOrPlot(TemplatePlot):
    """
    Bitmap plot with oriented lines draws for every units, representing
    the most preferred orientations.
    Constructs a matrix of drawing directives displaying oriented lines
    in each unit, colored according to the order or preference, and selectivity
    This plot expects channels named "OrX" "SelX", with "X" the number
    ranking the preferred orientations.
    """
    unit_size       = param.Number(default=25,bounds=(9,None),doc="box size of a single unit")
    min_brightness  = param.Number(default=30,bounds=(0,50),doc="min brightness of lines")
    max_brightness  = param.Number(default=90,bounds=(50,100),doc="max brightness of lines")

    def __init__(self,channels,sheet_views,density,
                 plot_bounding_box,normalize,
                 range_=False,**params):
        super(MultiOrPlot,self).__init__(channels,sheet_views,density, 
                                   plot_bounding_box,normalize,**params)
        
        n       = len( channels.keys() )
        if density > 10:
            self.unit_size = int( density )

        # there should be an even number of channels
        if n % 2:
            self.debug('Empty plot.')
            return

        if ( self.unit_size % 2 ) == 0:
            self.unit_size = self.unit_size + 1
        
        n       = n / 2
        m       = []
        for i in range( n ):
            o           = self._get_matrix( "Or%d" % (i+1) )
            s           = self._get_matrix( "Sel%d" % (i+1) )
            if ( o==None or s==None ):
                self.debug('Empty plot.')
                return
            m.append( ( o, s ) )
                
        shape,box         = self._get_shape_and_box()                                 
        dm                = self.__make_lines_from_or_matrix( m, shape )
        box_size          = self.unit_size
        self.bitmap       = DrawBitmap( dm, box_size )
        self._orig_bitmap = self.bitmap
        

    def __vertices_from_or( self, o ):
        """
        help function for generating coordinates of line vertices
        from normalized orientation value.
        Return a list with two tuples, the coordinates of the segment with the
        given orientation, in the normalized range [ 0...1 ].
	Orientation is expected in range [ 0..pi ]. Space
        representation is in ordinary image convention: first coordinate is X,
        from left to right, second coordinate Y, from top to bottom.
        """

        s       = 0.5 * sin( o )
        c       = 0.5 * cos( o )
        return [ ( 0.5 - c, 0.5 + s ), ( 0.5 + c, 0.5 - s ) ]
        

    def __make_line_directive( self, os_list ):
        """
        help function for composing the list of line directives
        for a single unit.
        """

        d_hue   = 360 / len( os_list )
        hue     = 0
        p       = []
        n       = self.max_brightness - self.min_brightness
        for o,s in os_list:
            if s > 0.:
                f       = "hsl(%d,100%%,%2d%%)" % ( hue, max( self.min_brightness, n * ( 1. - s ) ) )
                p.append( { "line": [ o, { "fill": f } ] } )
            hue         = hue + d_hue
        
        return p


    def __make_lines_from_or_matrix( self, matrices, shape ):
        """ 
        return a matrix of line drawing directives for each unit, derived from
        the given list of tuples ( o, s ), where o is the orientation view and s
        is the selectivity. The list is ordered by the orientation preference.
        """

        vertices_from_or        = numpy.vectorize( self.__vertices_from_or, otypes=[numpy.object_] )
        mat_list                = []
        for o, s in matrices:
            a   = s.mean()
            d   = s.std()
            ad  = a + d
        if isinstance( ad, numpy.number ) and ad > 0:
            mat_list.append( ( vertices_from_or( o ), ( s - d ) / ad ) )

        lines   = numpy.empty( shape, numpy.object_ )
        for x in range( shape[ 0 ] ):
            for y in range( shape[ 1 ] ):
                os_list = []
                for o, s in mat_list:
                    os_list.append( ( o[ x, y ], s[ x, y ] ) )
                lines[ x, y ]   = self.__make_line_directive( os_list )

        return lines
